#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <setjmp.h>
#include <jpeglib.h>
#include <zlib.h>
#include <string.h>
#include "base_types.h"
#include "define.h"
#include "bitmap_def.h"
#include "config.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif

//#define PRINT_DG

static	struct jpeg_decompress_struct jpegObject;
static	struct jpeg_source_mgr jpegSourceManager;
//static unsigned char *inputData2;
static unsigned char *inputData;
static int haveTables = 0;

struct my_error_mgr {
    struct jpeg_error_mgr pub;	/* "public" fields */
    jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;
struct my_error_mgr	jpegErrorMgr;

METHODDEF(void) my_error_exit (j_common_ptr cInfo) {
    my_error_ptr myerr = (my_error_ptr) cInfo->err;

    (*cInfo->err->output_message) (cInfo);

    longjmp(myerr->setjmp_buffer, 1);
}
//
// Methods for Source data manager
static void initSource(struct jpeg_decompress_struct *cInfo) {
    cInfo->src->bytes_in_buffer = 0;
}

static boolean fillInputBuffer(struct jpeg_decompress_struct *cInfo) {
    cInfo->src->next_input_byte = inputData;
    cInfo->src->bytes_in_buffer = 1;
    inputData++;

    return 1;
}

static void skipInputData(struct jpeg_decompress_struct *cInfo, long count) {
    cInfo->src->bytes_in_buffer = 0;
    inputData += count;
}

static boolean resyncToRestart(struct jpeg_decompress_struct *cInfo, int desired) {
    return jpeg_resync_to_restart(cInfo, desired);
}

static void termSource(struct jpeg_decompress_struct *cInfo) {}
/////////////////////
static int buildJpegAlpha(struct DefineBitsJPEG *b, unsigned char *buffer) {
    z_stream	stream;
    int		status;
    unsigned char  *data;
#ifdef PRINT_DG

    printf("flash: loading jpeg (buildJpegAlpha)\n");
#endif

    data = (unsigned char*)malloc(b->Width*b->Height);
    assert(data != NULL) ;

    stream.next_in = buffer;
    stream.avail_in = 1;
    stream.next_out = data;
    stream.avail_out = b->Width*b->Height;
    stream.zalloc = Z_NULL;
    stream.zfree = Z_NULL;

    status = inflateInit(&stream);

    while (1) {
        status = inflate(&stream, Z_SYNC_FLUSH) ;
        if (status == Z_STREAM_END) {
            break;
        }
        if (status != Z_OK) {
            printf("Zlib data error : %s\n", stream.msg);
            free(data);
            assert(0);
        }
        stream.avail_in = 1;
    }

    inflateEnd(&stream);

    b->Alpha_bufP = data;
#ifdef PRINT_DG

    printf("flash: loading jpeg (buildJpegAlpha) finished !\n");
#endif

    return 0;
}

int buildFromJpegInterchangeData (int read_alpha, long offset,struct DefineBitsJPEG* outStruct) {

    struct jpeg_decompress_struct cInfo;
    struct jpeg_source_mgr mySrcMgr;
    struct my_error_mgr jerr;
    JSAMPROW buffer[1];		/* Output row buffer */
    int row_stride;		/* physical row width in output buffer */
    unsigned char *ptrPix;
    long n;
    long		 nbColors;
    int status;
    //	struct DefineBitMap *outStruct;
#ifdef PRINT_DG

    printf("flash: loading jpeg (Interchange)\n");
#endif
    // Kludge to correct some corrupted files
    if (outStruct->JPEGData[1] == 0xd9 && outStruct->JPEGData[3] == 0xd8) {
        outStruct->JPEGData[3] = 0xd9;
        outStruct->JPEGData[1] = 0xd8;
    }

    cInfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = my_error_exit;
    /* Establish the setjmp return context for my_error_exit to use. */
    if (setjmp(jerr.setjmp_buffer)) {
        jpeg_destroy_decompress(&cInfo);
        if (outStruct->PixelsP) {
            free(outStruct->PixelsP);
            outStruct->PixelsP = NULL;
        }
        assert(0);
    }
    // Set current stream pointer to stream
    inputData = outStruct->JPEGData;

    jpeg_create_decompress(&cInfo);

    // Setup source manager structure
    mySrcMgr.init_source = initSource;
    mySrcMgr.fill_input_buffer = fillInputBuffer;
    mySrcMgr.skip_input_data = skipInputData;
    mySrcMgr.resync_to_restart = resyncToRestart;
    mySrcMgr.term_source = termSource;
    mySrcMgr.next_input_byte = NULL; /* until buffer loaded */

    // Set default source manager
    cInfo.src = &mySrcMgr;

    status = jpeg_read_header(&cInfo, FALSE);
    assert(status==JPEG_HEADER_OK||status==JPEG_HEADER_TABLES_ONLY);
#ifdef PRINT_DG

    printf("flash: loading jpeg (Interchange) jpeg_read_header(&cInfo, FALSE) :%d \n",status);
#endif


    status = jpeg_read_header(&cInfo, TRUE);
    assert(status==JPEG_HEADER_OK||status==JPEG_HEADER_TABLES_ONLY);
#ifdef PRINT_DG

    printf("flash: loading jpeg (Interchange) jpeg_read_header(&cInfo, TRUE) :%d \n",status);
#endif

    cInfo.quantize_colors = TRUE;	// Create colormapped image
    status = jpeg_start_decompress(&cInfo);
    assert(status==TRUE);
#ifdef PRINT_DG

    printf("flash: loading jpeg (Interchange) jpeg_start_decompress :%d \n",status);
#endif

    // Set objet dimensions
    outStruct->Height = cInfo.output_height;
    outStruct->Width = cInfo.output_width;
    outStruct->Bpl = outStruct->Width;


    outStruct->PixelsP = (unsigned char *)malloc(outStruct->Height*outStruct->Width);
#ifdef PRINT_DG

    printf("flash: loading jpeg (Interchange) outStruct->PixelsP beginPos :%d \n",outStruct->PixelsP);
    printf("flash: loading jpeg (Interchange) outStruct->Height :%d \n",outStruct->Height);
    printf("flash: loading jpeg (Interchange) outStruct->Width :%d \n",outStruct->Width);
    printf("flash: loading jpeg (Interchange) outStruct->output_components :%d \n",cInfo.output_components);
    printf("flash: loading jpeg (Interchange) length :%d \n",outStruct->Height*outStruct->Width);
    printf("flash: loading jpeg (Interchange) outStruct->PixelsP endPos :%d \n",outStruct->PixelsP + (outStruct->Height*outStruct->Width));
#endif

    assert(outStruct->PixelsP != NULL) ;

    ptrPix = outStruct->PixelsP;
    /* JSAMPLEs per row in output buffer */
    row_stride = cInfo.output_width * cInfo.output_components;
    /* Make a one-row-high sample array that will go away when done with image */
    buffer[0] = (JSAMPROW)malloc(row_stride);
    /*           jpeg_read_scanlines(...); */

    while (cInfo.output_scanline < cInfo.output_height) {
        status = jpeg_read_scanlines(&cInfo, buffer, 1);
        assert(status==1);
        memcpy(ptrPix,buffer[0],row_stride);
        ptrPix+= row_stride;
    }
#ifdef PRINT_DG
    printf("flash: loading jpeg (Interchange) ptrPix endPos :%d \n",ptrPix);
    printf("flash: loading jpeg (Interchange) remain space  :%d \n",outStruct->PixelsP+(outStruct->Height*outStruct->Width*2)-ptrPix);
#endif


    free(&(buffer[0][0]));

    outStruct->ColormapP = (struct RGBA*)malloc(cInfo.actual_number_of_colors*sizeof(struct RGBA));
    assert(outStruct->ColormapP != NULL) ;

    nbColors = cInfo.actual_number_of_colors;
    outStruct->Colors = nbColors;

    for(n=0; n < nbColors; n++) {
        outStruct->ColormapP[n].Red = cInfo.colormap[0][n];
        outStruct->ColormapP[n].Green = cInfo.colormap[1][n];
        outStruct->ColormapP[n].Blue = cInfo.colormap[2][n];
    }

    status = jpeg_finish_decompress(&cInfo);
    assert(status==TRUE);
    jpeg_destroy_decompress(&cInfo);

    if (read_alpha) {
        if (buildJpegAlpha(outStruct,  outStruct->JPEGData + offset) < 0) {
            assert(0);
        }
    }

#ifdef PRINT_DG
    printf("flash: loading jpeg (Interchange) finished!\n");
#endif

    return 0;
}



// Read JPEG image using pre-loaded Tables

int buildFromJpegAbbreviatedData(unsigned char *stream,struct DefineBits* outStruct) {
    //	struct jpeg_decompress_struct jpegObject;
    //	struct jpeg_source_mgr mySrcMgr;
    //	struct my_error_mgr jerr;
    JSAMPROW buffer[1];
    unsigned char *ptrPix;
    int stride;
    long n;
    long		 nbColors;
    int status;

#ifdef PRINT_DG

    printf("flash: loading jpeg (abbreviated)\n");
#endif

    // Set current stream pointer to stream
    inputData = stream;

    if (setjmp(jpegErrorMgr.setjmp_buffer)) {
        // JPEG data Error
        //		jpeg_destroy_decompress(&jpegObject);
        if (outStruct->PixelsP) {
            free(outStruct->PixelsP);
            outStruct->PixelsP = NULL;
        }
        assert(0);
    }

    status = jpeg_read_header(&jpegObject, TRUE);
    assert(status==JPEG_HEADER_OK||status==JPEG_HEADER_TABLES_ONLY);
#ifdef PRINT_DG

    printf("flash: loading jpeg (abbreviated) jpeg_read_header(&jpegObject, TRUE) :%d \n",status);
#endif

    jpegObject.quantize_colors = TRUE;	// Create colormapped image
    status = jpeg_start_decompress(&jpegObject);
    assert(status==TRUE);
#ifdef PRINT_DG

    printf("flash: loading jpeg (abbreviated) jpeg_start_decompress :%d \n",status);
#endif
#ifdef PRINT_DG

    printf("flash: loading jpeg (abbreviated) cInfo.desired_number_of_colors :%d \n",jpegObject.desired_number_of_colors);
    printf("flash: loading jpeg (abbreviated) cInfo.num_components :%d \n",jpegObject.num_components);
    printf("flash: loading jpeg (abbreviated) cInfo.jpeg_color_space :%d \n",jpegObject.jpeg_color_space);
    printf("flash: loading jpeg (abbreviated) cInfo.out_color_space :%d \n",jpegObject.out_color_space);
    printf("flash: loading jpeg (abbreviated) cInfo.out_color_components :%d \n",jpegObject.out_color_components);
    printf("flash: loading jpeg (abbreviated) cInfo.output_components :%d \n",jpegObject.output_components);
    printf("flash: loading jpeg (abbreviated) cInfo.actual_number_of_colors :%d \n",jpegObject.actual_number_of_colors);
    printf("flash: loading jpeg (abbreviated) cInfo.data_precision :%d \n",jpegObject.data_precision);
#endif

    // Set objet dimensions
    outStruct->Height = jpegObject.output_height;
    outStruct->Width = jpegObject.output_width;
    outStruct->Bpl = outStruct->Width;
    outStruct->PixelsP = (unsigned char *)malloc(outStruct->Height*outStruct->Width);
    assert(outStruct->PixelsP != NULL) ;
    ptrPix = outStruct->PixelsP;

    stride = jpegObject.output_width * jpegObject.output_components;

    buffer[0] = (JSAMPROW)malloc(stride);

    while (jpegObject.output_scanline < jpegObject.output_height) {
        status = jpeg_read_scanlines(&jpegObject, buffer, 1);
        assert(status==1);
        memcpy(ptrPix,buffer[0],stride);
        ptrPix+= stride;
    }

    free(&(buffer[0][0]));

    outStruct->ColormapP = (struct RGBA*)malloc(jpegObject.actual_number_of_colors*sizeof(struct RGBA));
    assert(outStruct->ColormapP != NULL) ;
    nbColors = jpegObject.actual_number_of_colors;
    outStruct->Colors = nbColors;

    for(n=0; n < nbColors; n++) {
        outStruct->ColormapP[n].Red = jpegObject.colormap[0][n];
        outStruct->ColormapP[n].Green = jpegObject.colormap[1][n];
        outStruct->ColormapP[n].Blue = jpegObject.colormap[2][n];
    }

    status = jpeg_finish_decompress(&jpegObject);
    assert(status==TRUE);

#ifdef PRINT_DG

    printf("flash: loading jpeg (abbreviated) jpeg_finish_decompress :%d \n",status);
    printf("flash: loading jpeg (abbreviated) finished!\n");
#endif

    return 0;
}

// Just init JPEG object and read JPEG Tables
int readJpegTables(struct JPEGTable *JTable) {
    //	struct my_error_mgr jerr;
    int rVal;
#ifdef PRINT_DG

    printf("flash: readJpegTables !\n");
#endif
    /**/
    if (haveTables) {
        //it has already been initialized
        return 0;
    }
    /**/
    // Setup error handler
    jpegObject.err = jpeg_std_error(&jpegErrorMgr.pub);
    jpegErrorMgr.pub.error_exit = my_error_exit;

    if (setjmp(jpegErrorMgr.setjmp_buffer)) {
        // JPEG data Error
        jpeg_destroy_decompress(&jpegObject);
        assert(0);
    }

    // Set current stream pointer to stream
    inputData = JTable->JPEGTables;

    // Here it's Ok

    jpeg_create_decompress(&jpegObject);

    // Setup source manager structure
    jpegSourceManager.init_source = initSource;
    jpegSourceManager.fill_input_buffer = fillInputBuffer;
    jpegSourceManager.skip_input_data = skipInputData;
    jpegSourceManager.resync_to_restart = resyncToRestart;
    jpegSourceManager.term_source = termSource;
    jpegSourceManager.next_input_byte = NULL; /* until buffer loaded */

    // Set default source manager
    jpegObject.src = &jpegSourceManager;

    rVal = jpeg_read_header(&jpegObject, FALSE);
    assert(rVal ==JPEG_HEADER_OK||rVal ==JPEG_HEADER_TABLES_ONLY);
#ifdef PRINT_DG

    printf("flash: readJpegTables jpeg_read_header : %d\n",rVal);
#endif

    haveTables = 1;

    return 0;
}

int buildFromZlibData(unsigned char *buffer, int width, int height, int format, int tableSize, int tableHasAlpha,struct DefineBitsLossless2* outStruct) {
    z_stream	stream;
    int		status;
    unsigned char  *data;
    int		elementSize;
    long		 depth=0;
    long		 nbColors=0;

#ifdef PRINT_DG

    printf("flash: loading with zlib\n");
#endif

    if (tableHasAlpha) {
        elementSize = 4;	// Cmap is RGBA
    } else {
        elementSize = 3;	// Cmap is RGB
    }

    stream.next_in = buffer;
    stream.avail_in = 1;
    stream.zalloc = Z_NULL;
    stream.zfree = Z_NULL;

    tableSize++;

    outStruct->Width = width;
    outStruct->Height = height;
    outStruct->Bpl = width;
    // Uncompress struct RGBA Table
    if (format == 3) {
        unsigned char *colorTable;
        long n;

        // Ajust width for 32 bit padding
        width = (width+3)/4*4;
        outStruct->Width = width;
        outStruct->Bpl = width;

        depth = 1;
        colorTable = (unsigned char*)malloc(tableSize*elementSize);
        assert(colorTable != NULL);

        stream.next_out = colorTable;
        stream.avail_out = tableSize*elementSize;

        inflateInit(&stream);

        while (1) {
            status = inflate(&stream, Z_SYNC_FLUSH);
            if (status == Z_STREAM_END) {
                break;
            }
            if (status != Z_OK) {
                printf("Zlib cmap error : %s\n", stream.msg);
                assert(0);
            }
            stream.avail_in = 1;
            // Colormap if full
            if (stream.avail_out == 0) {
                break;
            }
        }

        nbColors = tableSize;

        outStruct->ColormapP = (struct RGBA*)malloc(nbColors*sizeof(struct RGBA));
        if (outStruct->ColormapP == NULL) {
            free(colorTable);
            assert(0);
        }

        for(n=0; n < nbColors; n++) {
            outStruct->ColormapP[n].Red = colorTable[n*elementSize+0];
            outStruct->ColormapP[n].Green = colorTable[n*elementSize+1];
            outStruct->ColormapP[n].Blue = colorTable[n*elementSize+2];
            if (tableHasAlpha) {
                outStruct->ColormapP[n].Alpha = colorTable[n*elementSize+3];
            }
        }

        free(colorTable);

    } else if (format == 4) {
        depth = 2;
        width = (width+1)/2*2;
        outStruct->Bpl = width;
    } else if (format == 5) {
        depth = 4;
    }

    data = (unsigned char*)malloc(depth*width*height);
    if (data == NULL) {
        if (outStruct->ColormapP)
            free(outStruct->ColormapP);
        assert(0);
    }

    stream.next_out = data;
    stream.avail_out = depth*width*height;

    if (format != 3) {
        status = inflateInit(&stream);
    }

    while (1) {
        status = inflate(&stream, Z_SYNC_FLUSH) ;
        if (status == Z_STREAM_END) {
            break;
        }
        if (status != Z_OK) {
            printf("Zlib data error : %s\n", stream.msg);
            free(data);
            assert(0);
        }
        stream.avail_in = 1;
    }

    inflateEnd(&stream);

    outStruct->PixelsP = (unsigned char *)malloc(height*width);
    if (outStruct->PixelsP == NULL) {
        if (outStruct->ColormapP)
            free(outStruct->ColormapP);
        free(data);
        assert(0);
    }

    if (format != 3) {
        int n,c;
        unsigned char r,g,b,a;
        unsigned char *ptr;

        r = g = b = a = 0; /* to supress warnings */

        nbColors = 0;
        outStruct->ColormapP = (struct RGBA*)malloc(256*sizeof(struct RGBA));
        if (outStruct->ColormapP == NULL) {
            free(data);
            free(outStruct->PixelsP);
            assert(0);
        }
        memset(outStruct->ColormapP, 0, 256 * sizeof(struct RGBA));
        ptr = outStruct->PixelsP;

        for(n=0; n < width*height*depth; n+=depth,ptr++) {

            switch (format) {
            case 4:
                a = 1;
                r = (data[n] & 0x78)<<1;
                g = ((data[n] & 0x03)<<6) | (data[n+1] & 0xc0)>>2;
                b = (data[n+1] & 0x1e)<<3;
                break;
            case 5:
                a = data[n];
                // Reduce color dynamic range
                r = data[n+1]&0xe0;
                g = data[n+2]&0xe0;
                b = data[n+3]&0xe0;
                break;
            }
            for(c=0; c < nbColors; c++) {
                if (r == outStruct->ColormapP[c].Red
                        &&  g == outStruct->ColormapP[c].Green
                        &&  b == outStruct->ColormapP[c].Blue) {
                    *ptr = c;
                    break;
                }
            }
            if (c == nbColors) {
                if (nbColors == 256)
                    continue;
                nbColors++;
                if (nbColors == 256) {
                    //printf("Colormap entries exhausted. After %d scanned pixels\n", n/4);
                }
                outStruct->ColormapP[c].Alpha = a;
                outStruct->ColormapP[c].Red   = r;
                outStruct->ColormapP[c].Green = g;
                outStruct->ColormapP[c].Blue  = b;
                *ptr = c;
            }
        }
    } else {
        memcpy(outStruct->PixelsP, data, width*height);
        if (tableHasAlpha) {
            int n;
            unsigned char *ptr, *alpha;

            outStruct->Alpha_bufP = (unsigned char *)malloc(width*height);
            ptr = data;
            alpha = outStruct->Alpha_bufP;
            for(n=0; n < width*height; n++, ptr++, alpha++) {
                *alpha = outStruct->ColormapP[*ptr].Alpha;
            }
        }
    }
    outStruct->Colors = nbColors;

    assert(outStruct->PixelsP != 0);
    free(data);
#ifdef PRINT_DG

    printf("flash: loading with zlib finished!\n");
#endif

    return 0;
}




int bmp_makebmp(struct DefineCharacter* InStruct,swf_file_t *swf_p) {
    struct DefineBits* InBitsStruct;
    struct DefineBitsJPEG* InJPEGStruct;
    struct DefineBitsLossless2* InBitsLossless2Struct;
    int rc = 0;
#ifdef PRINT_DG

    printf("flash: enter BitmapCache !\n");
    printf("flash: InStruct->CharacterType : %d\n",InStruct->CharacterType);
#endif

    if(InStruct->CharacterType == (UI8)CharBitsOne) {
        InBitsStruct = (struct DefineBits*)InStruct;

        readJpegTables(swf_p->JPEGTableP);
#ifdef PRINT_DG

        printf("flash: InBitsStruct->JPEGDataLength : %d\n",InBitsStruct->JPEGDataLength);
#endif

        rc = buildFromJpegAbbreviatedData(&InBitsStruct->JPEGData[0],InBitsStruct);
        if(rc) {
            assert(0);
        }

    } else if(InStruct->CharacterType ==(UI8)CharBitsJPEG) {
        InJPEGStruct = (struct DefineBitsJPEG*)InStruct;
#ifdef PRINT_DG

        printf("flash: InJPEGStruct->JPEGDataLength : %d\n",InJPEGStruct->JPEGDataLength);
        printf("flash: InJPEGStruct->AlphaDataLength : %d\n",InJPEGStruct->AlphaDataLength);
        /*
        	for(j=0;j<20;j++)
        		printf(" %x ",InJPEGStruct->JPEGData[j]);
        	printf("\n");
        	for(i=InJPEGStruct->JPEGDataLength-20;i<InJPEGStruct->JPEGDataLength;i++)
        		printf(" %x ",InJPEGStruct->JPEGData[i]);
        	printf("\n");
        */
#endif

        if(InJPEGStruct->AlphaDataLength) {
            rc = buildFromJpegInterchangeData(1,InJPEGStruct->JPEGDataLength,InJPEGStruct);
        } else {
            rc = buildFromJpegInterchangeData(0,0,InJPEGStruct);
        }
        if(rc) {
            assert(0);
        }
    } else if(InStruct->CharacterType == (UI8)CharBitsLosslessOne) {
        InBitsLossless2Struct = (struct DefineBitsLossless2*)InStruct;
#ifdef PRINT_DG

        printf("flash: InBitsLossless2Struct->ZlibBitmapDataLength : %d\n",InBitsLossless2Struct->ZlibBitmapDataLength);
#endif

        rc = buildFromZlibData(InBitsLossless2Struct->ZlibBitmapData,
                               InBitsLossless2Struct->BitmapWidth,InBitsLossless2Struct->BitmapHeight, InBitsLossless2Struct->BitmapFormat,
                               InBitsLossless2Struct->BitmapColorTableSize, 0,InBitsLossless2Struct);
        if(rc) {
            assert(0);
        }
    } else if(InStruct->CharacterType == (UI8)CharBitsLosslessTwo) {
        InBitsLossless2Struct = (struct DefineBitsLossless2*)InStruct;
#ifdef PRINT_DG

        printf("flash: InBitsLossless2Struct->ZlibBitmapDataLength : %d\n",InBitsLossless2Struct->ZlibBitmapDataLength);
#endif

        rc = buildFromZlibData(InBitsLossless2Struct->ZlibBitmapData,
                               InBitsLossless2Struct->BitmapWidth,InBitsLossless2Struct->BitmapHeight, InBitsLossless2Struct->BitmapFormat,
                               InBitsLossless2Struct->BitmapColorTableSize, 1,InBitsLossless2Struct);
        if(rc) {
            assert(0);
        }
    }

    assert(((struct DefineBits*)(InStruct))->PixelsP != 0);
#ifdef PRINT_DG

    printf("flash: enter BitmapCache finished !\n");
#endif

    return 0;
}
