#include <math.h>
#include <stdio.h>
#include <assert.h>
#include "config.h"
#include "base_types.h"
#include "define.h"
#include "movie_clips_def.h"
#include "swf_file_pub.h"
#include "morphing_ins.h"
#include "morphing_def.h"
#include <stdlib.h>
#include <glib.h>
#if WITH_DMALLOC
#include <dmalloc.h>
#endif

//#define PRINT_LOG
//#define PRINT_DG

//get Current RGBA
int CurrRGBA(GNode* node_p,struct RGBA *Start,struct RGBA *End,struct RGBA *Curr) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);
    Curr->Red	=	Start->Red+(End->Red-Start->Red)*data_p->ratio/65535;
    Curr->Green	=	Start->Green+(End->Green-Start->Green)*data_p->ratio/65535;
    Curr->Blue	=	Start->Blue+(End->Blue-Start->Blue)*data_p->ratio/65535;
    Curr->Alpha	=	Start->Alpha+(End->Alpha-Start->Alpha)*data_p->ratio/65535;
    Curr->pixel	=	Start->pixel+(End->pixel-Start->pixel)*data_p->ratio/65535;
    return 0;
}
//get Current Matrix
int CurrMatrix(GNode* node_p,struct Matrix *Start,struct Matrix *End,struct Matrix *Curr) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);
    Curr->HasScale = Start->HasScale;
    Curr->HasRotate = Start->HasRotate;

    Curr->ScaleX	=	Start->ScaleX+(End->ScaleX-Start->ScaleX)*data_p->ratio/65535;
    Curr->ScaleY	=	Start->ScaleY+(End->ScaleY-Start->ScaleY)*data_p->ratio/65535;
    Curr->RotateSkew0	=	Start->RotateSkew0+(End->RotateSkew0-Start->RotateSkew0)*data_p->ratio/65535;
    Curr->RotateSkew1	=	Start->RotateSkew1+(End->RotateSkew1-Start->RotateSkew1)*data_p->ratio/65535;
    Curr->TranslateX	=	Start->TranslateX+(End->TranslateX-Start->TranslateX)*data_p->ratio/65535;
    Curr->TranslateY	=	Start->TranslateY+(End->TranslateY-Start->TranslateY)*data_p->ratio/65535;
    return 0;
}
//get Current MorphGradRecord
int CurrGradient(GNode* node_p,struct MorphGradRecord *Morph,struct Gradient **Curr) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);
    struct MorphGradRecord *tmpMorph;
    struct Gradient *Grad;
    int i = 0;

    tmpMorph = Morph;
    while(tmpMorph) {
        tmpMorph = tmpMorph->NextP;
        i++;
    }
    if(i==0) {
        Grad=(struct Gradient *)malloc(sizeof(struct Gradient)+(sizeof(struct GradRecord)*i));
        bzero((void *)(Grad),sizeof(struct Gradient)+(sizeof(struct GradRecord)*i));
    } else {
        Grad=(struct Gradient *)malloc(sizeof(struct Gradient)+(sizeof(struct GradRecord)*(i-1)));
        bzero((void *)(Grad),sizeof(struct Gradient)+(sizeof(struct GradRecord)*(i-1)));
    }
    assert(Grad != NULL);
    Grad->NumGradients = i;
    i = 0;
    while(Morph) {
        Grad->GradientRecords[i].Ratio = Morph->StartRatio+(Morph->EndRatio-Morph->StartRatio)*data_p->ratio/65535;
        CurrRGBA(node_p,&Morph->StartColor,&Morph->EndColor,&(Grad->GradientRecords[i].Color));
        Morph = Morph->NextP;
        i++;
    }

    *Curr = Grad;
    return 0;
}

//get Current FillStyle
int CurrFillStyle(GNode* node_p,struct MorphFillStyle *Morph,struct FillStyle *Curr) {
    Curr->FillStyleType = Morph->FillStyleType;
    //RGBA Morph
    CurrRGBA(node_p,&Morph->StartColor,&Morph->EndColor,&Curr->Color);
    //GradientMatrix Morph
    CurrMatrix(node_p,&Morph->StartGradientMatrix,&Morph->EndGradientMatrix,&Curr->GradientMatrix);
    //MorphGradRecord morph
    CurrGradient(node_p,Morph->GradientP,&Curr->GradientP);
    Curr->BitmapId = Morph->BitmapId;
    //BitmapMatrix morph
    CurrMatrix(node_p,&Morph->StartBitmapMatrix,&Morph->EndBitmapMatrix,&Curr->BitmapMatrix);

    return 0;
}
//deal with MorphFillStyleArray morhp
int CurrFillArray(GNode* node_p,struct MorphFillStyleArray *Morph,struct FillStyleArray **Curr) {
    int i;
    struct FillStyleArray *tmpFill;
    if(Morph->count==0) {
        tmpFill=(struct FillStyleArray*)malloc(sizeof(struct FillStyleArray)+sizeof(struct FillStyle)*(Morph->count));
        bzero((void *)(tmpFill),(sizeof(struct FillStyleArray)+sizeof(struct FillStyle)*(Morph->count)));
    } else {
        tmpFill=(struct FillStyleArray*)malloc(sizeof(struct FillStyleArray)+sizeof(struct FillStyle)*(Morph->count-1));
        bzero((void *)(tmpFill),(sizeof(struct FillStyleArray)+sizeof(struct FillStyle)*(Morph->count-1)));
    }
    assert(tmpFill != NULL);
    for(i=0;i<Morph->count;i++) {
        CurrFillStyle(node_p,&Morph->MorphFillStyles[i],&tmpFill->FillStyles[i]);
    }
    tmpFill->count = Morph->count ;
    *Curr = tmpFill;
    return 0;
}
//deal with MorphLineStyle morph
int CurrLineStyle(GNode* node_p,struct MorphLineStyle *Morph,struct LineStyle *Curr) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);

    Curr->Width	=	Morph->StartWidth+(Morph->EndWidth-Morph->StartWidth)*data_p->ratio/65535;
    //RGBA morph
    CurrRGBA(node_p,&Morph->StartColor,&Morph->EndColor,&Curr->ColorRGB);

    return 0;
}
//deal with MorphLineStyleArray morph
int CurrLineArray(GNode* node_p,struct MorphLineStyleArray *Morph,struct LineStyleArray **Curr) {
    int i;
    struct LineStyleArray *tmpFill;
    if(Morph->count==0) {
        tmpFill=(struct LineStyleArray*)malloc(sizeof(struct LineStyleArray)+sizeof(struct LineStyle)*(Morph->count));
        bzero((void *)(tmpFill),(sizeof(struct LineStyleArray)+sizeof(struct LineStyle)*(Morph->count)));
    } else {
        tmpFill=(struct LineStyleArray*)malloc(sizeof(struct LineStyleArray)+sizeof(struct LineStyle)*(Morph->count-1));
        bzero((void *)(tmpFill),(sizeof(struct LineStyleArray)+sizeof(struct LineStyle)*(Morph->count-1)));
    }
    assert(tmpFill != NULL);
    for(i=0;i<Morph->count;i++) {
        CurrLineStyle(node_p,&Morph->MorphLineStyles[i],&tmpFill->LineStyles[i]);
    }
    tmpFill->count = Morph->count ;
    *Curr = tmpFill;
    return 0;
}

//deal with StyleChangeRecord
int StartToCurrent(GNode* node_p,struct StyleChangeRecord* StartRed,struct StyleChangeRecord* EndRed,struct StyleChangeRecord* outRed) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);

    outRed->MoveDeltaX = StartRed->MoveDeltaX + (EndRed->MoveDeltaX-StartRed->MoveDeltaX)*data_p->ratio/65535;
    outRed->MoveDeltaY = StartRed->MoveDeltaY + (EndRed->MoveDeltaY-StartRed->MoveDeltaY)*data_p->ratio/65535;

    return 0;
}
//curved morph
int CurvedToCurved(GNode* node_p,struct CurvedEdgeRecord* SCurRed,struct CurvedEdgeRecord* ECurRed,struct CurvedEdgeRecord* outCurRed) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);

    outCurRed->ControlDeltaX = SCurRed->ControlDeltaX +(ECurRed->ControlDeltaX-SCurRed->ControlDeltaX)*data_p->ratio/65535;
    outCurRed->ControlDeltaY = SCurRed->ControlDeltaY +(ECurRed->ControlDeltaY-SCurRed->ControlDeltaY)*data_p->ratio/65535;
    outCurRed->AnchorDeltaX = SCurRed->AnchorDeltaX +(ECurRed->AnchorDeltaX-SCurRed->AnchorDeltaX)*data_p->ratio/65535;
    outCurRed->AnchorDeltaY = SCurRed->AnchorDeltaY +(ECurRed->AnchorDeltaY-SCurRed->AnchorDeltaY)*data_p->ratio/65535;

    return 0;
}

//Straight morph
int StraightToStraight(GNode* node_p,struct StraightEdgeRecord* SRed,struct StraightEdgeRecord* ERed,struct StraightEdgeRecord* outRed) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);

    if(SRed->StraightLineFlag) {
        if(ERed->StraightLineFlag) {
            outRed->StraightLineFlag = 1;
            outRed->StraightLineDeltaX = SRed->StraightLineDeltaX +(ERed->StraightLineDeltaX-SRed->StraightLineDeltaX)*data_p->ratio/65535;
            outRed->StraightLineDeltaY = SRed->StraightLineDeltaY +(ERed->StraightLineDeltaY-SRed->StraightLineDeltaY)*data_p->ratio/65535;
            outRed->VertLineFlag = 0;
            outRed->HorLineDeltaX = 0;
            outRed->VertLineDeltaY = 0;
        } else if(ERed->VertLineFlag) {
            outRed->StraightLineFlag = 1;
            outRed->StraightLineDeltaX = SRed->StraightLineDeltaX +(0-SRed->StraightLineDeltaX)*data_p->ratio/65535;
            outRed->StraightLineDeltaY = SRed->StraightLineDeltaY +(ERed->VertLineDeltaY-SRed->StraightLineDeltaY)*data_p->ratio/65535;
            outRed->VertLineFlag = 0;
            outRed->HorLineDeltaX = 0;
            outRed->VertLineDeltaY = 0;
        } else {
            outRed->StraightLineFlag = 1;
            outRed->StraightLineDeltaX = SRed->StraightLineDeltaX +(ERed->HorLineDeltaX-SRed->StraightLineDeltaX)*data_p->ratio/65535;
            outRed->StraightLineDeltaY = SRed->StraightLineDeltaY +(0-SRed->StraightLineDeltaY)*data_p->ratio/65535;
            outRed->VertLineFlag = 0;
            outRed->HorLineDeltaX = 0;
            outRed->VertLineDeltaY = 0;
        }
    } else if(SRed->VertLineFlag) {
        if(ERed->StraightLineFlag) {
            outRed->StraightLineFlag = 1;
            outRed->StraightLineDeltaX = 0 +(ERed->StraightLineDeltaX-0)*data_p->ratio/65535;
            outRed->StraightLineDeltaY = SRed->VertLineDeltaY +(ERed->StraightLineDeltaY-SRed->VertLineDeltaY)*data_p->ratio/65535;
            outRed->VertLineFlag = 0;
            outRed->HorLineDeltaX = 0;
            outRed->VertLineDeltaY = 0;
        } else if(ERed->VertLineFlag) {
            outRed->StraightLineFlag = 0;
            outRed->StraightLineDeltaX = 0;
            outRed->StraightLineDeltaY = 0;
            outRed->VertLineFlag = 1;
            outRed->HorLineDeltaX = 0;
            outRed->VertLineDeltaY = SRed->VertLineDeltaY +(ERed->VertLineDeltaY-SRed->VertLineDeltaY)*data_p->ratio/65535;
        } else {
            outRed->StraightLineFlag = 1;
            outRed->StraightLineDeltaX = 0 +(ERed->HorLineDeltaX-0)*data_p->ratio/65535;
            outRed->StraightLineDeltaY = SRed->VertLineDeltaY +(0-SRed->VertLineDeltaY)*data_p->ratio/65535;
            outRed->VertLineFlag = 0;
            outRed->HorLineDeltaX = 0;
            outRed->VertLineDeltaY = 0;
        }
    } else {
        if(ERed->StraightLineFlag) {
            outRed->StraightLineFlag = 1;
            outRed->StraightLineDeltaX = SRed->HorLineDeltaX +(ERed->StraightLineDeltaX-SRed->HorLineDeltaX)*data_p->ratio/65535;
            outRed->StraightLineDeltaY = 0 +(ERed->StraightLineDeltaY-0)*data_p->ratio/65535;
            outRed->VertLineFlag = 0;
            outRed->HorLineDeltaX = 0;
            outRed->VertLineDeltaY = 0;
        } else if(ERed->VertLineFlag) {
            outRed->StraightLineFlag = 1;
            outRed->StraightLineDeltaX = SRed->HorLineDeltaX +(0-SRed->HorLineDeltaX)*data_p->ratio/65535;
            outRed->StraightLineDeltaY = 0+(ERed->VertLineDeltaY-0)*data_p->ratio/65535;
            outRed->VertLineFlag = 0;
            outRed->HorLineDeltaX = 0;
            outRed->VertLineDeltaY = 0;
        } else {
            outRed->StraightLineFlag = 0;
            outRed->StraightLineDeltaX = 0;
            outRed->StraightLineDeltaY =0;
            outRed->VertLineFlag = 0;
            outRed->HorLineDeltaX = SRed->HorLineDeltaX +(ERed->HorLineDeltaX-SRed->HorLineDeltaX)*data_p->ratio/65535;
            outRed->VertLineDeltaY = 0;
        }
    }

    return 0;
}

//verify the straight to curved
int VerifyStraight(GNode* node_p,struct StraightEdgeRecord* StrRed,struct CurvedEdgeRecord* EndRed,struct CurvedEdgeRecord* CurRed) {
    double pf;
    double a,b,c,d;
    a = (EndRed->ControlDeltaX)*(EndRed->ControlDeltaX);
    b = (EndRed->ControlDeltaY)*(EndRed->ControlDeltaY);
    c = (EndRed->AnchorDeltaX-EndRed->ControlDeltaX)*(EndRed->AnchorDeltaX-EndRed->ControlDeltaX);
    d = (EndRed->AnchorDeltaY-EndRed->ControlDeltaY)*(EndRed->AnchorDeltaY-EndRed->ControlDeltaY);
    pf = sqrt(a+b)/(sqrt(a+b)+sqrt(c+d));

    if(StrRed->StraightLineFlag) {
        CurRed->ControlDeltaX = (StrRed->StraightLineDeltaX)*pf;
        CurRed->ControlDeltaY = (StrRed->StraightLineDeltaY)*pf;
        CurRed->AnchorDeltaX = StrRed->StraightLineDeltaX - CurRed->ControlDeltaX;
        CurRed->AnchorDeltaY = StrRed->StraightLineDeltaY - CurRed->ControlDeltaY;
    } else if(StrRed->VertLineFlag) {
        CurRed->ControlDeltaX = 0;
        CurRed->ControlDeltaY = (StrRed->VertLineDeltaY)*pf;
        CurRed->AnchorDeltaX = 0;
        CurRed->AnchorDeltaY = StrRed->VertLineDeltaY - CurRed->ControlDeltaY;
    } else {
        CurRed->ControlDeltaX = (StrRed->HorLineDeltaX)*pf;
        CurRed->ControlDeltaY = 0;
        CurRed->AnchorDeltaX = StrRed->HorLineDeltaX - CurRed->ControlDeltaX;
        CurRed->AnchorDeltaY = 0;
    }
    return 0;
}

//change straight to curved
int StraightToCurved(struct StraightEdgeRecord* StrRed,struct CurvedEdgeRecord* CurRed) {

    if(StrRed->StraightLineFlag) {
        CurRed->ControlDeltaX = StrRed->StraightLineDeltaX/2;
        CurRed->ControlDeltaY = StrRed->StraightLineDeltaY/2;
        CurRed->AnchorDeltaX = StrRed->StraightLineDeltaX/2;
        CurRed->AnchorDeltaY = StrRed->StraightLineDeltaY/2;
    } else if(StrRed->VertLineFlag) {
        CurRed->ControlDeltaX = 0;
        CurRed->ControlDeltaY = StrRed->VertLineDeltaY/2;
        CurRed->AnchorDeltaX = 0;
        CurRed->AnchorDeltaY = StrRed->VertLineDeltaY/2;
    } else {
        CurRed->ControlDeltaX = StrRed->HorLineDeltaX/2;
        CurRed->ControlDeltaY = 0;
        CurRed->AnchorDeltaX = StrRed->HorLineDeltaX/2;
        CurRed->AnchorDeltaY = 0;
    }

    return 0;
}

//get current RECT Bounds
int CurrBounds(GNode* node_p,struct RECT StartBounds,struct RECT EndBounds,struct RECT *Curr) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);

    Curr->Xmin=StartBounds.Xmin+(EndBounds.Xmin-StartBounds.Xmin)*data_p->ratio/65535;
    Curr->Xmax=StartBounds.Xmax+(EndBounds.Xmax-StartBounds.Xmax)*data_p->ratio/65535;
    Curr->Ymin=StartBounds.Ymin+(EndBounds.Ymin-StartBounds.Ymin)*data_p->ratio/65535;
    Curr->Ymax=StartBounds.Ymax+(EndBounds.Ymax-StartBounds.Ymax)*data_p->ratio/65535;

    return 0;
}

//deal with Shape morph
int GetShapePos(GNode* node_p,struct ShapeRecord* StartRed,struct ShapeRecord* EndRed,struct ShapeRecord** outCurRed) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);
    struct ShapeRecord* tmpCurRed;
    struct StraightEdgeRecord* StartStraightRecord = NULL;
    struct CurvedEdgeRecord* StartCurvedRecord;
    struct StyleChangeRecord* StartSChangeRecord;

    struct StraightEdgeRecord* EndStraightRecord = NULL;
    struct CurvedEdgeRecord* EndCurvedRecord;
    struct StyleChangeRecord* EndSChangeRecord;

    struct StraightEdgeRecord* outStraightRecord;
    struct CurvedEdgeRecord* outCurvedRecord;
    struct StyleChangeRecord* outSChangeRecord;
#ifdef PRINT_DG

    printf("flash: loading GetShapePos Start---\n");
#endif
    //judge shape
    if(StartRed->Flags&128) {
        //start record
        if(StartRed->Flags&64) {
            //straight record
            StartCurvedRecord = (struct CurvedEdgeRecord*)malloc(sizeof(struct CurvedEdgeRecord));
            assert(StartCurvedRecord != NULL) ;
            bzero((void *)(StartCurvedRecord),sizeof(struct CurvedEdgeRecord));
            StartStraightRecord = (struct StraightEdgeRecord*)StartRed;
            if(EndRed->Flags&64) {
                data_p->isStraight = 0;
                StraightToCurved(StartStraightRecord,StartCurvedRecord);
            } else {
                data_p->isStraight = 0;
                EndCurvedRecord = (struct CurvedEdgeRecord*)EndRed;
                VerifyStraight(node_p,StartStraightRecord,EndCurvedRecord,StartCurvedRecord);
            }
        } else {
            //curved record
            data_p->isStraight = 0;
            StartCurvedRecord = (struct CurvedEdgeRecord*)StartRed;
        }

        //end record
        if(EndRed->Flags&64) {
            //straight record
            EndCurvedRecord= (struct CurvedEdgeRecord*)malloc(sizeof(struct CurvedEdgeRecord));
            assert(EndCurvedRecord!= NULL);
            bzero((void *)(EndCurvedRecord),sizeof(struct CurvedEdgeRecord));
            EndStraightRecord = (struct StraightEdgeRecord*)EndRed;
            if(StartRed->Flags&64) {
                data_p->isStraight = 0;
                StraightToCurved(EndStraightRecord,EndCurvedRecord);
            } else {
                StartCurvedRecord = (struct CurvedEdgeRecord*)StartRed;
                VerifyStraight(node_p,EndStraightRecord,StartCurvedRecord,EndCurvedRecord);
            }
        } else {
            //curved record
            EndCurvedRecord = (struct CurvedEdgeRecord*)EndRed;
        }
        //do morph
        if(data_p->isStraight == 1) {
            outStraightRecord = (struct StraightEdgeRecord*)malloc(sizeof(struct StraightEdgeRecord));
            assert(outStraightRecord != NULL);
            bzero((void *)(outStraightRecord),sizeof(struct StraightEdgeRecord));
            StraightToStraight(node_p,StartStraightRecord,EndStraightRecord,outStraightRecord);

            tmpCurRed = (struct ShapeRecord*)outStraightRecord;
            tmpCurRed->Flags = StartRed->Flags;

        } else {
            outCurvedRecord = (struct CurvedEdgeRecord*)malloc(sizeof(struct CurvedEdgeRecord));
            assert(outCurvedRecord != NULL);
            bzero((void *)(outCurvedRecord),sizeof(struct CurvedEdgeRecord));
            CurvedToCurved(node_p,StartCurvedRecord,EndCurvedRecord,outCurvedRecord);

            tmpCurRed = (struct ShapeRecord*)outCurvedRecord;
            tmpCurRed->Flags = StartRed->Flags&191;
        }
    } else {
        StartSChangeRecord = (struct StyleChangeRecord*)StartRed;
        EndSChangeRecord = (struct StyleChangeRecord*)EndRed;
        outSChangeRecord = (struct StyleChangeRecord*)malloc(sizeof(struct StyleChangeRecord));
        assert(outSChangeRecord != NULL);
        bzero((void *)(outSChangeRecord),sizeof(struct StyleChangeRecord));
        if(StartRed->Flags&1) {
            //start & end point morph
            StartToCurrent(node_p,StartSChangeRecord,EndSChangeRecord,outSChangeRecord);
        }
        outSChangeRecord->FillStyle0 = StartSChangeRecord->FillStyle0;
        outSChangeRecord->FillStyle1 = StartSChangeRecord->FillStyle1;
        outSChangeRecord->LineStyle = StartSChangeRecord->LineStyle;
        outSChangeRecord->FillStylesP = StartSChangeRecord->FillStylesP;
        outSChangeRecord->LineStylesP = StartSChangeRecord->LineStylesP;
        outSChangeRecord->NumFillBits = StartSChangeRecord->NumFillBits;
        outSChangeRecord->NumLineBits = StartSChangeRecord->NumLineBits;

        tmpCurRed = (struct ShapeRecord*)outSChangeRecord;
        tmpCurRed->Flags = StartRed->Flags;
    }

    *outCurRed = tmpCurRed;
    /*
    	if(StartRed->Flags&64)
    		if(StartCurvedRecord)
    			free(StartCurvedRecord );
    	if(EndRed->Flags&64)
    		if(EndCurvedRecord)
    			free(EndCurvedRecord );
    */
#ifdef PRINT_DG

    printf("flash: loading GetShapePos End---\n");
#endif

    return 0;
}
int freeShape(GNode* node_p) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);

    struct ShapeRecord *tempShape;
    free(data_p->GlobalShape.FillStylesP);
    data_p->GlobalShape.FillStylesP = NULL;
    free(data_p->GlobalShape.LineStylesP);
    data_p->GlobalShape.LineStylesP = NULL;
    tempShape = NULL;
    if(data_p->GlobalShape.ShapeRecordsP)
        while(data_p->GlobalShape.ShapeRecordsP) {
            tempShape =data_p->GlobalShape.ShapeRecordsP->NextP;
            free(data_p->GlobalShape.ShapeRecordsP);
            data_p->GlobalShape.ShapeRecordsP = tempShape;
        }
    /*
    	if(&data_p->GlobalShape)
    		freeDefineShape(&data_p->GlobalShape);
    */
    return 0;
}
int mallocShape(GNode* node_p) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);

    data_p->GlobalShape.FillStylesP = (struct FillStyleArray*)malloc(sizeof(struct FillStyleArray));
    assert(data_p->GlobalShape.FillStylesP != NULL);
    bzero((void *)(data_p->GlobalShape.FillStylesP),sizeof(struct FillStyleArray));
    data_p->GlobalShape.LineStylesP = (struct LineStyleArray*)malloc(sizeof(struct LineStyleArray));
    assert(data_p->GlobalShape.LineStylesP != NULL);
    bzero((void *)(data_p->GlobalShape.LineStylesP),sizeof(struct LineStyleArray));
    data_p->GlobalShape.ShapeRecordsP = (struct ShapeRecord*)malloc(sizeof(struct ShapeRecord));
    assert(data_p->GlobalShape.ShapeRecordsP != NULL);
    bzero((void *)(data_p->GlobalShape.ShapeRecordsP),sizeof(struct ShapeRecord));
    return 0;
}
//add Flags
int AddFlags(struct  DefineMorphShape *Morph,struct ShapeRecord *FlagRed) {
    struct StyleChangeRecord	*SChangeRecord;
    struct StyleChangeRecord	*CurFlag;

    SChangeRecord = (struct StyleChangeRecord*)Morph->StartEdgesP;
    CurFlag = (struct StyleChangeRecord*)FlagRed;
    SChangeRecord->Flags = SChangeRecord->Flags|CurFlag->Flags;
    //SChangeRecord->Flags = CurFlag->Flags;
    SChangeRecord->MoveDeltaX = CurFlag->MoveDeltaX ;
    SChangeRecord->MoveDeltaY = CurFlag->MoveDeltaY ;
    SChangeRecord->LineStyle = CurFlag->LineStyle ;
    SChangeRecord->FillStylesP = CurFlag->FillStylesP;
    SChangeRecord->LineStylesP = CurFlag->LineStylesP;
    SChangeRecord->NumLineBits= CurFlag->NumLineBits;

    return 0;
}
//delete multy edges for out shape to solid fill
int DelEdges(GNode* node_p,struct  DefineMorphShape *Morph,struct ShapeRecord *FlagRed) {
    mrp_ins_t	*data_p = (mrp_ins_t*)(node_p->data);
    struct StyleChangeRecord	*SChangeRecord;
    struct StyleChangeRecord	*CurFlag;
    struct StyleChangeRecord	*CurEndFlag;
    struct ShapeRecord	*tmpRecord;

    if(Morph->StartEdgesP->Flags&1)
        return 0;

    if(Morph->StartEdgesP->Flags&6) {
        SChangeRecord = (struct StyleChangeRecord*)Morph->StartEdgesP;
        CurFlag = (struct StyleChangeRecord*)FlagRed;
        CurFlag->Flags = SChangeRecord->Flags|(CurFlag->Flags&9);
        CurFlag->FillStyle0 = SChangeRecord->FillStyle0;
        CurFlag->FillStyle1 = SChangeRecord->FillStyle1;
        CurFlag->NumFillBits = SChangeRecord->NumFillBits;
    }
    Morph->StartEdgesP = Morph->StartEdgesP->NextP ;
    if(Morph->EndEdgesP->Flags&1) {
        Morph->EndEdgesP = Morph->EndEdgesP->NextP ;
    }

    while(Morph->StartEdgesP) {
        if(Morph->StartEdgesP->Flags&1) {

            if(Morph->EndEdgesP->Flags&1) {
                break;
            } else {
                tmpRecord = Morph->EndEdgesP;
                CurEndFlag =(struct StyleChangeRecord*)malloc(sizeof(struct StyleChangeRecord));
                bzero((void *)(CurEndFlag),sizeof(struct StyleChangeRecord));
                CurEndFlag->Flags = 1;

                CurEndFlag->MoveDeltaX = data_p->EndFlagP.MoveDeltaX;
                CurEndFlag->MoveDeltaY = data_p->EndFlagP.MoveDeltaY;

                CurEndFlag->NextP = tmpRecord;
                Morph->EndEdgesP = (struct ShapeRecord*)CurEndFlag;

                break;
            }
        }
        if(Morph->StartEdgesP->Flags&6) {
            if(Morph->EndEdgesP->Flags&1) {
                //				CurEndFlag = (struct StyleChangeRecord *)Morph->EndEdgesP;
                //				FlagRed = Morph->StartEdgesP;
                AddFlags(Morph,FlagRed);
                break;
            }
            //				Morph->EndEdgesP = Morph->EndEdgesP->NextP ;
            Morph->StartEdgesP = Morph->StartEdgesP->NextP ;
        } else {
            if(Morph->EndEdgesP->Flags&1) {
                CurEndFlag = (struct StyleChangeRecord *)Morph->EndEdgesP;
                Morph->EndEdgesP = Morph->EndEdgesP->NextP ;
            } else {
                Morph->StartEdgesP = Morph->StartEdgesP->NextP ;
                Morph->EndEdgesP = Morph->EndEdgesP->NextP ;
            }
        }
    }
    return 0;
}
int AddFrame(struct  DefineMorphShape *Morph) {
    struct StyleChangeRecord	*StrRecord;
    struct StyleChangeRecord	*EndRecord;
    struct ShapeRecord	*tmpRecord;

    StrRecord = (struct StyleChangeRecord*)Morph->StartEdgesP;
    StrRecord->Flags = StrRecord->Flags|1;
    StrRecord->MoveDeltaX = 0;
    StrRecord->MoveDeltaY = 0;

    tmpRecord = Morph->EndEdgesP;
    if(tmpRecord->Flags&1) {
        return 0;
    }

    EndRecord = (struct StyleChangeRecord*)malloc(sizeof(struct StyleChangeRecord));
    bzero((void *)(EndRecord),sizeof(struct StyleChangeRecord));
    EndRecord->Flags = 1;
    EndRecord->MoveDeltaX = 0;
    EndRecord->MoveDeltaY = 0;
    EndRecord->FillStyle0 = 0;
    EndRecord->FillStyle1 = 0;
    EndRecord->LineStyle = 0;
    EndRecord->FillStylesP = NULL;
    EndRecord->LineStylesP = NULL;
    EndRecord->NumLineBits = 0;
    EndRecord->NumFillBits = 0;

    EndRecord->NextP = tmpRecord;
    Morph->EndEdgesP = (struct ShapeRecord*)EndRecord;

    return 0;
}

int setStyleChangeRecord(struct StyleChangeRecord *red,struct ShapeRecord *ShapeP) {
    struct StyleChangeRecord *tmpRed;

    tmpRed = (struct StyleChangeRecord*)ShapeP;

    red->Flags = tmpRed->Flags;
    red->MoveDeltaX = tmpRed->MoveDeltaX;
    red->MoveDeltaY = tmpRed->MoveDeltaY;
    red->FillStyle0 = tmpRed->FillStyle0;
    red->FillStyle1 = tmpRed->FillStyle1;
    red->LineStyle = tmpRed->LineStyle;
    red->FillStylesP = tmpRed->FillStylesP;
    red->LineStylesP = tmpRed->LineStylesP;
    red->NumLineBits = tmpRed->NumLineBits;
    red->NumFillBits = tmpRed->NumFillBits;

    return 0;
}
//the func MorphShape change to Shape
void mrp_ins_update_shape(GNode* node_p) {
    mrp_ins_t	 		*data_p = 	(mrp_ins_t*)(node_p->data);
    struct DefineMorphShape 	*inMorphShape = (struct DefineMorphShape*)(data_p->define_p);
    //	struct DefineShape	*outShape =

    struct ShapeRecord *headShapeP;
    struct ShapeRecord *SShapeP,*RShapeP;
    struct  DefineMorphShape tmpMorphShape;

    struct ShapeRecord *FlagP = NULL;

#ifdef PRINT_LOG

    printf("flash: loading MorphShapeToShap Start---\n");
#endif

//    if (data_p->last_ratio == data_p->ratio)
//        return;

    freeShape(node_p);
    mallocShape(node_p);
    tmpMorphShape = *inMorphShape;
    //deal with RECT
    CurrBounds(node_p,tmpMorphShape.StartBounds,tmpMorphShape.EndBounds,&data_p->GlobalShape.ShapeBounds);
    //deal with ShapeRecord
    //	data_p->EndFlagP.MoveDeltaX = 0;
    //	data_p->EndFlagP.MoveDeltaY = 0;
    if(!(tmpMorphShape.StartEdgesP->Flags&1)) {
        AddFrame(&tmpMorphShape);
    }
    headShapeP = NULL;
    RShapeP = NULL;
    while(tmpMorphShape.StartEdgesP) {
        if(tmpMorphShape.StartEdgesP->Flags&6) {
            DelEdges(node_p,&tmpMorphShape,FlagP);
            //			AddFlags(&tmpMorphShape,FlagP);
            if(tmpMorphShape.StartEdgesP==NULL)
                break;
        }
        SShapeP = (struct ShapeRecord*)malloc(sizeof(struct ShapeRecord));
        assert(SShapeP != NULL);
        bzero((void *)(SShapeP),sizeof(struct ShapeRecord));
        GetShapePos(node_p,tmpMorphShape.StartEdgesP,tmpMorphShape.EndEdgesP,&SShapeP);
        //		SShapeP->Flags = tmpMorphShape.StartEdgesP->Flags ;
        if(headShapeP==NULL)
            headShapeP = SShapeP;
        else
            RShapeP->NextP = SShapeP;

        RShapeP = SShapeP;
        if(tmpMorphShape.StartEdgesP->Flags&128) {
            tmpMorphShape.StartEdgesP = tmpMorphShape.StartEdgesP->NextP;
            tmpMorphShape.EndEdgesP = tmpMorphShape.EndEdgesP->NextP;
        } else {
            if(tmpMorphShape.StartEdgesP->Flags&1) {
                FlagP = SShapeP;
                tmpMorphShape.StartEdgesP = tmpMorphShape.StartEdgesP->NextP;
                if(tmpMorphShape.EndEdgesP->Flags&1) {
                    setStyleChangeRecord(&data_p->EndFlagP,tmpMorphShape.EndEdgesP) ;
                    tmpMorphShape.EndEdgesP = tmpMorphShape.EndEdgesP->NextP;
                }
            } else {
                tmpMorphShape.StartEdgesP = tmpMorphShape.StartEdgesP->NextP;
                if(tmpMorphShape.EndEdgesP->Flags&1) {
                    tmpMorphShape.EndEdgesP = tmpMorphShape.EndEdgesP->NextP;
                }

            }
        }
    }
    if(RShapeP!=NULL)
        RShapeP->NextP = NULL;
    data_p->GlobalShape.ShapeRecordsP = headShapeP;

    //deal with MorphFillStyleArray
    CurrFillArray(node_p,tmpMorphShape.MorphFillStylesP,&data_p->GlobalShape.FillStylesP);
    //deal with MorphLineStyleArray
    CurrLineArray(node_p,tmpMorphShape.MorphLineStylesP,&data_p->GlobalShape.LineStylesP);
    data_p->GlobalShape.CharacterType = CharShape;
    data_p->shape_p = &data_p->GlobalShape;

//    data_p->last_ratio = data_p->ratio;
#ifdef PRINT_LOG

    printf("flash: loading MorphShapeToShap End---\n");
    printf("--------flash: loading MorphShapeToShap data_p->ratio--------------%u \n",data_p->ratio);
    if(Ratio == 0)
        printf("--------flash: loading MorphShapeToShap data_p->ratio--------------\n");
#endif

}

void mrp_ins_change(GNode* node_p,struct PlaceObject *cmd_p) {
    mrp_ins_t    *data_p   = (mrp_ins_t*)(node_p->data);
    if (cmd_p->Flags & PlaceFlagHasRatio) {
        data_p->ratio = cmd_p->Ratio;
    };
    if (data_p->last_ratio != data_p->ratio) {
        mrp_ins_update_shape(node_p);
	data_p->pic_dirty = TRUE;
    };
    data_p->last_ratio = data_p->ratio;
};
