#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"
#include "spliter.h"
#include "spliter_pub.h"
#include "spliter_reader.h"
#include "base_types.h"
#include "bitmap_def.h"
#include "button_def.h"
#include "dynamic_text_def.h"
#include "font_def.h"
#include "morphing_def.h"
#include "movie_clips_def.h"
#include "shape_def.h"
#include "static_text_def.h"
#include "sound_def.h"
#include "define.h"
#include "swf_file_pub.h"
#include <glib.h>
#include <assert.h>
#if WITH_DMALLOC
#include <dmalloc.h>
#endif

int gg_error(void) {
    printf("read file error!\n");
    return -1;
};

int g_readdata(struct reader_t *reader,int len) {
    void *p;
    if (len == 0)
        return 0;
    p=(void *)malloc(len);
    if (reader->read(reader, p, len) !=len )
        return gg_error();
    free((void *)p);
    return 0;
}

SI32 reader_GetSBits(struct reader_t *reader, int nbits) {
    UI32 res = reader_readbits(reader, nbits);
    if (res&(1<<(nbits-1)))
        res|=(0xffffffff<<nbits);
    return (SI32)res;
}

UI32 reader_GetBits(struct reader_t *reader, int nbits) {
    return reader_readbits(reader, nbits);
}

int reader_GetString(struct reader_t *reader,char **str) {
    char tmpstr[1024];
    int i=0;
    char *cp;
    char p;
    do {
        if ((reader->read(reader, &p, 1) != 1))
            return gg_error();
        tmpstr[i]=p;
        i++;
    } while(p);
    cp=malloc(i);
    memcpy(cp,tmpstr,i);
    *str=cp;
    return i;
}

int reader_GetStringWithLen(struct reader_t *reader,char **str,int len) {
    char *cp;
    cp=malloc(len+1);
    if ((reader->read(reader, cp, len) != len))
        return gg_error();
    *str=cp;
    cp+=len;
    *cp=0;
    return 0;
}

int reader_GetRgb(struct reader_t *reader,struct RGBA * r) {
    UI8 p[3];
    if (reader->read(reader, p, 3) != 3)
        return gg_error();
    r->Red=p[0];
    r->Green=p[1];
    r->Blue=p[2];
    r->Alpha=255;
    return 0;
}

int reader_GetRgba(struct reader_t *reader,struct RGBA * r) {
    UI8 p[4];
    if (reader->read(reader, p, 4) != 4)
        return gg_error();
    r->Red=p[0];
    r->Green=p[1];
    r->Blue=p[2];
    r->Alpha=p[3];
    return 0;
}

int reader_GetRect(struct reader_t *reader,struct RECT * r) {
    int nbits;
    struct RECT dummy;
    reader->bitpos=8;
    if (!r)
        r = &dummy;
    nbits = (int) reader_GetBits(reader,5);
    r->Xmin = reader_GetSBits(reader,nbits);
    r->Xmax = reader_GetSBits(reader,nbits);
    r->Ymin = reader_GetSBits(reader,nbits);
    r->Ymax = reader_GetSBits(reader,nbits);
    return 0;
}

int reader_GetMatrix(struct reader_t *reader,struct Matrix *p) {
    int nbits;
    reader->bitpos=8;
    p->HasScale=(UI8)reader_GetBits(reader,1);
    if (p->HasScale) {
        nbits=(int)reader_GetBits(reader,5);
        p->ScaleX=(double)reader_GetSBits(reader,nbits)/(double)0x10000;
        p->ScaleY=(double)reader_GetSBits(reader,nbits)/(double)0x10000;
    } else {
        p->ScaleX=1;
        p->ScaleY=1;
    }
    p->HasRotate=(UI8)reader_GetBits(reader,1);
    if (p->HasRotate) {
        nbits=(int)reader_GetBits(reader,5);
        p->RotateSkew1=(double)reader_GetSBits(reader,nbits)/(double)0x10000;
        p->RotateSkew0=(double)reader_GetSBits(reader,nbits)/(double)0x10000;
    }
    nbits=(int)reader_GetBits(reader,5);
    p->TranslateX=reader_GetSBits(reader,nbits);
    p->TranslateY=reader_GetSBits(reader,nbits);
    return 0;
}

int reader_GetCxFromWithAlpha(struct reader_t *reader,struct CxFormWithAlpha *p) {
    int nbits;
    reader->bitpos=8;
    p->HasAddTerms=(UI8)reader_GetBits(reader,1);
    p->HasMultTerms=(UI8)reader_GetBits(reader,1);
    nbits=(int)reader_GetBits(reader,4);
    if (p->HasMultTerms) {
        p->RedMultTerm=(FB)reader_GetSBits(reader,nbits)/(FB)0x100;
        p->GreenMultTerm=(FB)reader_GetSBits(reader,nbits)/(FB)0x100;
        p->BlueMultTerm=(FB)reader_GetSBits(reader,nbits)/(FB)0x100;
        p->AlphaMultTerm=(FB)reader_GetSBits(reader,nbits)/(FB)0x100;
    } else {
        p->RedMultTerm=1;
        p->GreenMultTerm=1;
        p->BlueMultTerm=1;
        p->AlphaMultTerm=1;
    }
    if (p->HasAddTerms) {
        p->RedAddTerm=reader_GetSBits(reader,nbits);
        p->GreenAddTerm=reader_GetSBits(reader,nbits);
        p->BlueAddTerm=reader_GetSBits(reader,nbits);
        p->AlphaAddTerm=reader_GetSBits(reader,nbits);
    }
    return 0;
}

int reader_GetCxFrom(struct reader_t *reader,struct CxFormWithAlpha *p) {
    int nbits;
    reader->bitpos=8;
    p->HasAddTerms=(UI8)reader_GetBits(reader,1);
    p->HasMultTerms=(UI8)reader_GetBits(reader,1);
    nbits=(int)reader_GetBits(reader,4);
    if (p->HasMultTerms) {
        p->RedMultTerm=(FB)reader_GetSBits(reader,nbits)/(FB)0x100;
        p->GreenMultTerm=(FB)reader_GetSBits(reader,nbits)/(FB)0x100;
        p->BlueMultTerm=(FB)reader_GetSBits(reader,nbits)/(FB)0x100;
        p->AlphaMultTerm=0x100;
    }
    if (p->HasAddTerms) {
        p->RedAddTerm=reader_GetSBits(reader,nbits);
        p->GreenAddTerm=reader_GetSBits(reader,nbits);
        p->BlueAddTerm=reader_GetSBits(reader,nbits);
        p->AlphaAddTerm=0;
    }
    return 0;
}

int reader_GetGradient(struct reader_t *reader,struct Gradient **p,int type) {
    struct Gradient *Grad;
    UI8 Num;
    int i;
    if (reader->read(reader, &(Num), 1) !=1 )
        return gg_error();
    Grad=(struct Gradient *)malloc(sizeof(struct Gradient)+(sizeof(struct GradRecord)*(Num-1)));
    bzero((void *)Grad,sizeof(struct Gradient)+(sizeof(struct GradRecord)*(Num-1)));
    Grad->NumGradients=Num;
    for (i=0;i<Num;i++) {
        if (reader->read(reader, &(Grad->GradientRecords[i].Ratio), 1) !=1 )
            return gg_error();
        if (type==3) {
            reader_GetRgba(reader,&(Grad->GradientRecords[i].Color));
        } else {
            reader_GetRgb(reader,&(Grad->GradientRecords[i].Color));
        }
    }
    *p=Grad;
    return 0;
}

int g_GetColor(struct reader_t*reader,struct RGB *color) {
    UI8 p[3];
    if (reader->read(reader, p, 3) != 3)
        return gg_error();
    color->Red=p[0];
    color->Green=p[1];
    color->Blue=p[2];
    return 0;
}

int g_AddToDict(swf_file_t *swf_p,int id,struct DefineCharacter* char_p) {

    if (swf_p->dict_p == NULL) {
        swf_p->dict_p = g_new0(struct Dict,1);
        swf_p->dict_p->count = 0;
    };
    if (swf_p->dict_p->count <= id) {
        swf_p->dict_p->count = id + 50;
        swf_p->dict_p = g_realloc(swf_p->dict_p,sizeof(struct Dict)+(swf_p->dict_p->count)*4);
    };
    swf_p->dict_p->dict[id] = char_p;
    char_p->CharacterID = id;
    switch (char_p->CharacterType) {
    case CharBitsOne             :
    case CharBitsJPEG            :
    case CharBitsLosslessOne     :
    case CharBitsLosslessTwo     :
        bmp_makebmp(char_p,swf_p);
        break;
    case CharText:
        stx_to_shape(swf_p,(struct DefineText*)char_p);
        break;
    };
    return 0;
}

int g_GetDefSound(struct reader_t *reader,swf_file_t *swf_p,int sound_len) {
    struct DefineSound *sound_p;
    int i;
    UI16        SoundId;
    UI8		SoundFlags;
    UI32        SoundSampleCount;
    UI32	SoundLength;
    if (reader->read(reader, &SoundId, 2) !=2 )
        return gg_error();
    SoundId=SWAP16(SoundId);
    if (reader->read(reader, &SoundFlags, 1) !=1 )
        return gg_error();
    if (reader->read(reader, &SoundSampleCount, 4) !=4 )
        return gg_error();
    SoundSampleCount=SWAP32(SoundSampleCount);
    SoundLength=sound_len-7;
    i=sizeof(struct DefineSound)+sound_len-1;
    sound_p=(struct DefineSound *)malloc(i);
    sound_p->CharacterType=CharEventSound;
    sound_p->SoundId=SoundId;
    sound_p->SoundFlags=SoundFlags;
    sound_p->SoundSampleCount=SoundSampleCount;
    sound_p->SoundLength=SoundLength;
    if (reader->read(reader, sound_p->SoundData, SoundLength) !=SoundLength )
        return gg_error();
    return g_AddToDict(swf_p,SoundId,(struct DefineCharacter*)sound_p);
}
//g_getExportAssets is not compilete
int g_GetExportAssets(struct reader_t *reader,swf_file_t *swf_p,int len) {
    UI16	count,id;
    STRING	name_p;
    int		i;
    struct asset 	*asset_p,*a_p;

    if (reader->read(reader, &count, 2) !=2 )
        return gg_error();
    count=SWAP16(count);

    for (i = 0; i< count ; i++) {
        if (reader->read(reader, &id, 2) !=2 )
            return gg_error();
        id=SWAP16(id);

        reader_GetString(reader,&(name_p));

        asset_p 	= g_new0(struct asset,1);
        asset_p->id 	= id;
        asset_p->name_p = name_p;

        a_p = swf_p->dict_p->assets_p;
        if (a_p) {
            while (a_p->next_p)
                a_p=a_p->next_p;
            a_p->next_p=asset_p;
        } else {
            swf_p->dict_p->assets_p=asset_p;
        }
        printf("export %d name %s \n",id, name_p);
    };
    return 0;
};

int g_GetFillStyle(struct reader_t *reader,struct FillStyle *FillStyle_p,int type) {
    struct FillStyle *p;
    p=FillStyle_p;
    if (reader->read(reader, &(p->FillStyleType), 1) !=1 )
        return gg_error();
    if (p->FillStyleType==0x00) {
        if(type==3) {
            reader_GetRgba(reader,&(p->Color));
        } else {
            reader_GetRgb(reader,(&(p->Color)));
        }
    }
    if ((p->FillStyleType==0x10)||(p->FillStyleType==0x12)) {
        reader_GetMatrix(reader,&(p->GradientMatrix));
        reader_GetGradient(reader,&(p->GradientP),type);
    }
    if ((p->FillStyleType==0x41)||(p->FillStyleType==0x42)||(p->FillStyleType==0x43)||(p->FillStyleType==0x40)) {
        if (reader->read(reader, &p->BitmapId, 2) !=2 )
            return gg_error();
        p->BitmapId=SWAP16(p->BitmapId);
        reader_GetMatrix(reader,&(p->BitmapMatrix));
    }
    return 0;
}

int g_GetFillStyleArray(struct reader_t *reader,struct FillStyleArray **p,int type) {
    UI8 count;
    UI16 e_count;
    int len;
    int i;
#if DEBUG

    printf("begin GetFillStyleArray function\n");
#endif

    if (reader->read(reader, &count, 1) !=1 )
        return gg_error();
    len=count;
    if (type!=1) {
        if (count==0xff) {
            if (reader->read(reader, &e_count, 2) !=2 )
                return gg_error();
            e_count=SWAP16(e_count);
            len=e_count;
        }
    }
#if DEBUG
    printf("len:%d\n",len);
#endif

    *p=(struct FillStyleArray*)malloc(sizeof(struct FillStyleArray)+sizeof(struct FillStyle)*(len-1));
    bzero((void *)*p,sizeof(struct FillStyleArray)+sizeof(struct FillStyle)*(len-1));
    (*p)->count=len;
    for(i=0;i<len;i++) {
#if DEBUG
        printf("get the %d FillStyles:%p\n",i,&((*p)->FillStyles[i]));
#endif

        g_GetFillStyle(reader,(struct FillStyle *)&((*p)->FillStyles[i]),type);
    }
#if DEBUG
    printf("end GetFillStyleArray function\n");
#endif

    return 0;
}

int g_GetLineStyle(struct reader_t *reader,struct LineStyle *LineStyle_p,int type) {
    struct LineStyle *p;
    p=LineStyle_p;
    bzero((void *)p,sizeof(struct LineStyle));
    if (reader->read(reader, &(p->Width), 2) !=2 )
        return gg_error();
    p->Width=SWAP16(p->Width);
    if (type==3) {
        reader_GetRgba(reader,&(p->ColorRGB));
    } else {
        reader_GetRgb(reader,(&(p->ColorRGB)));
    }
    return 0;
}

int g_GetLineStyleArray(struct reader_t *reader,struct LineStyleArray **p,int type) {
    UI8 count;
    UI16 e_count;
    int len;
    int i;
    if (reader->read(reader, &count, 1) !=1 )
        return gg_error();
    if (count==0xff) {
        if (reader->read(reader, &e_count, 2) !=2 )
            return gg_error();
        e_count=SWAP16(e_count);
        len=e_count;
    } else {
        len=count;
    }
    *p=(struct LineStyleArray*)malloc(sizeof(struct LineStyleArray)+sizeof(struct LineStyle)*(len-1));
    bzero((void *)*p,sizeof(struct LineStyleArray)+sizeof(struct LineStyle)*(len-1));
    (*p)->count=len;
    for(i=0;i<len;i++) {
        g_GetLineStyle(reader,&((*p)->LineStyles[i]),type);
    }
    return 0;
}

int g_GetShapeRecords(struct reader_t *reader,struct ShapeRecord **ShapeRecord_p,int type) {
    UI8 NumFillBits,NumLineBits;
    UI8 TypeFlag,Flags;
    int MoveBits;
    int NumBits;
    struct StraightEdgeRecord *StraightEdgeRecordP;
    struct CurvedEdgeRecord   *CurvedEdgeRecordP;
    struct StyleChangeRecord  *StyleChangeRecordP;
    struct ShapeRecord *ShapeRecordP,*dq = NULL;
    reader->bitpos=8;
    NumFillBits=reader_GetBits(reader,4);
    NumLineBits=reader_GetBits(reader,4);
    while (1) {
        TypeFlag=(UI8)reader_GetBits(reader,1);
        if (TypeFlag==0) {
            Flags=(UI8)reader_GetBits(reader,5);
            if (Flags==0) {
#if DEBUG
                printf("EndShapeRecord\n");
#endif

                return 0;
            } else {
#if DEBUG
                printf("StyleChangeRecord\n");
#endif

                StyleChangeRecordP=(struct StyleChangeRecord *)malloc(sizeof(struct StyleChangeRecord));
                bzero((void *)StyleChangeRecordP,sizeof(struct StyleChangeRecord));
                StyleChangeRecordP->Flags=Flags;
                if ((Flags&(StateMoveTo))==(StateMoveTo)) {
                    MoveBits=(int)reader_GetBits(reader,5);
                    StyleChangeRecordP->MoveDeltaX=(long)reader_GetSBits(reader,MoveBits);
                    StyleChangeRecordP->MoveDeltaY=(long)reader_GetSBits(reader,MoveBits);
                }
                if ((Flags&(StateFillStyle0))==(StateFillStyle0)) {
                    StyleChangeRecordP->FillStyle0=(long)reader_GetBits(reader,NumFillBits);
                }
                if ((Flags&(StateFillStyle1))==(StateFillStyle1)) {
                    StyleChangeRecordP->FillStyle1=(long)reader_GetBits(reader,NumFillBits);
                }
                if ((Flags&(StateLineStyle))==(StateLineStyle)) {
                    StyleChangeRecordP->LineStyle=(long)reader_GetBits(reader,NumLineBits);
                }
                if ((Flags&(StateNewStyles))==(StateNewStyles)) {
                    g_GetFillStyleArray(reader,&(StyleChangeRecordP->FillStylesP),type);
                    g_GetLineStyleArray(reader,&(StyleChangeRecordP->LineStylesP),type);
                    NumFillBits=(long)reader_GetBits(reader,4);
                    NumLineBits=(long)reader_GetBits(reader,4);
                }
                ShapeRecordP=(struct ShapeRecord *)StyleChangeRecordP;
            }
        } else {
            Flags=(UI8)reader_GetBits(reader,1);
            if (Flags==0) {
#if DEBUG
                printf("CurvedEdgeRecord\n");
#endif

                CurvedEdgeRecordP=(struct CurvedEdgeRecord *)malloc(sizeof(struct CurvedEdgeRecord));
                bzero((void *)CurvedEdgeRecordP,sizeof(struct CurvedEdgeRecord));
                CurvedEdgeRecordP->Flags=(TCurvedEdge);
                NumBits=(int)reader_GetBits(reader,4);
                NumBits+=2;
                CurvedEdgeRecordP->ControlDeltaX=(long)reader_GetSBits(reader,NumBits);
                CurvedEdgeRecordP->ControlDeltaY=(long)reader_GetSBits(reader,NumBits);
                CurvedEdgeRecordP->AnchorDeltaX=(long)reader_GetSBits(reader,NumBits);
                CurvedEdgeRecordP->AnchorDeltaY=(long)reader_GetSBits(reader,NumBits);
                ShapeRecordP=(struct ShapeRecord *)CurvedEdgeRecordP;
            } else {
#if DEBUG
                printf("StraightEdgeRecord\n");
#endif

                StraightEdgeRecordP=(struct StraightEdgeRecord *)malloc(sizeof(struct StraightEdgeRecord));
                bzero((void *)StraightEdgeRecordP,sizeof(struct StraightEdgeRecord));
                StraightEdgeRecordP->Flags=(TStraightEdge);
                NumBits=(int)reader_GetBits(reader,4);
                NumBits+=2;
                StraightEdgeRecordP->StraightLineFlag=(UB_1)reader_GetBits(reader,1);
                if (StraightEdgeRecordP->StraightLineFlag) {
                    StraightEdgeRecordP->StraightLineDeltaX=(long)reader_GetSBits(reader,NumBits);
                    StraightEdgeRecordP->StraightLineDeltaY=(long)reader_GetSBits(reader,NumBits);
                } else {
                    StraightEdgeRecordP->VertLineFlag=(UB_1)reader_GetBits(reader,1);
                    if (StraightEdgeRecordP->VertLineFlag) {
                        StraightEdgeRecordP->VertLineDeltaY=(long)reader_GetSBits(reader,NumBits);
                    } else {
                        StraightEdgeRecordP->HorLineDeltaX=(long)reader_GetSBits(reader,NumBits);
                    }
                }
                ShapeRecordP=(struct ShapeRecord *)StraightEdgeRecordP;
            }
        }
        if (*ShapeRecord_p==NULL) {
            *ShapeRecord_p=ShapeRecordP;
        } else {
            if ( dq == NULL) {
                dq=*ShapeRecord_p;
            }
            while(dq->NextP) {
                dq=dq->NextP;
            }
            dq->NextP=ShapeRecordP;
        }
    }
}

int g_GetShapeWithStyle(struct reader_t *reader,struct DefineShape *shape_p,int type) {
    g_GetFillStyleArray(reader,&(shape_p->FillStylesP),type);
    g_GetLineStyleArray(reader,&(shape_p->LineStylesP),type);
    g_GetShapeRecords(reader,&(shape_p->ShapeRecordsP),type);
    return 0;
}

int g_GetDefineShape(struct reader_t *reader,swf_file_t *swf_p,int len,int type) {
    struct DefineShape *p;
    UI16 id;
    int size;
    size=sizeof(struct DefineShape);
    p=(struct DefineShape *)malloc(size);
#if DEBUG

    printf("GetDefineShape function begin\n");
    printf("p:%p\n",p);
    printf("size%d\n",size);
#endif

    bzero((void*)p,sizeof(struct DefineShape));
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    id=SWAP16(id);
#if DEBUG

    printf("id:%d\n",id);
#endif

    p->CharacterType=CharShape;
#if DEBUG

    printf("p->CharacterType:%d\n",p->CharacterType);
#endif

    reader_GetRect(reader,&(p->ShapeBounds));
    g_GetShapeWithStyle(reader,p,type);
#if DEBUG

    printf("GetDefineShape function end\n");
#endif


    return g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
    return 0;
}

int g_GetDefineBits(struct reader_t *reader,swf_file_t *swf_p,int len) {
    struct DefineBits *p;
    UI16 id;
    int datacount=0;
    datacount=len-2;
    p=(struct DefineBits *)malloc(sizeof(struct DefineBits)+datacount);
    bzero((void *)p,sizeof(struct DefineBits)+datacount);
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    p->JPEGDataLength=datacount;
    p->CharacterType=(CharBitsOne);
    if (reader->read(reader, p->JPEGData, datacount) != datacount)
        return gg_error();
    //    bmp_makebmp(p);
    return g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
}

int g_GetDefineBitsJPEG(struct reader_t *reader,swf_file_t *swf_p,int len,int type) {
    struct DefineBitsJPEG *p;
    UI16 id;
    char cp[4];
    int datacount=0;
    if (type==3) {
        datacount=len-6;
    } else {
        datacount=len-2;
    }
    p=(struct DefineBitsJPEG *)malloc(sizeof(struct DefineBitsJPEG)+datacount);
    bzero((void *)p,sizeof(struct DefineBitsJPEG)+datacount);
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    p->CharacterType=(CharBitsJPEG);
    if (type==3) {
        if (reader->read(reader, cp, 4) !=4 )
            return gg_error();
        p->JPEGDataLength=GET32(cp);
        p->AlphaDataLength=datacount-p->JPEGDataLength;
    } else {
        p->JPEGDataLength=datacount;
    }
    if (reader->read(reader, p->JPEGData, datacount) != datacount)
        return gg_error();
    //    bmp_makebmp(p);
    return g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
}

int g_GetShowframe(struct reader_t *reader) {
    struct MovieClips *p;
    p=reader->MovieClipsPointer;
    while (p->NextP) {
        p=p->NextP;
    }
    p->frame+=1;
    return 0;
}

int g_GetSoundEnvelope(struct reader_t *reader,struct SoundEnvelope *EnvelopeP) {
    char strp[4];
    UI16 Num;
    if (reader->read(reader, (void*)strp, 4) !=4 )
        return gg_error();
    EnvelopeP->Pos44=GET32(strp);
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    EnvelopeP->LeftLevel=SWAP16(Num);
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    EnvelopeP->RightLevel=SWAP16(Num);
    return 0;
}

int g_GetSoundInfo(struct reader_t *reader,int len,struct SoundInfo **SoundInfoP) {
    struct SoundInfo* p;
    UI8 flags=0;
    char strp[4];
    UI32 InPoint=0,OutPoint=0;
    UI16 LoopCount=0;
    UI8 EnvPoints=0;
    int count=1;
    int i;
    if (reader->read(reader, &flags, 1) !=1 )
        return gg_error();
    if ((flags&HasInPoint)==HasInPoint) {
        if (reader->read(reader, (void*)strp, 4) !=4 )
            return gg_error();
        InPoint=GET32(strp);
    }
    if ((flags&HasOutPoint)==HasOutPoint) {
        if (reader->read(reader, (void*)strp, 4) !=4 )
            return gg_error();
        OutPoint=GET32(strp);
    }
    if ((flags&HasLoops)==HasLoops) {
        if (reader->read(reader, &LoopCount, 2) !=2 )
            return gg_error();
        LoopCount=SWAP16(LoopCount);
    }
    if ((flags&HasEnvelope)==HasEnvelope) {
        if (reader->read(reader, &EnvPoints, 1) !=1 )
            return gg_error();
        if (EnvPoints!=0) {
            count=EnvPoints;
        }
    }
    p=(struct SoundInfo*)malloc(sizeof(struct SoundInfo)+sizeof(struct SoundEnvelope)*(count-1));
    bzero((void*)p,(sizeof(struct SoundInfo)+sizeof(struct SoundEnvelope)*(count-1)));
    p->Flags=flags;
    p->InPoint=InPoint;
    p->OutPoint=OutPoint;
    p->LoopCount=LoopCount;
    p->EnvPoints=EnvPoints;
    for (i=0;i<EnvPoints;i++) {
        g_GetSoundEnvelope(reader,&(p->EnvelopeRecords[i]));
    }
    *SoundInfoP = p;
    return 0;
}

int g_GetStartSound(struct reader_t *reader,int len) {
    struct MovieClips *p;
    struct  StartSound *Sp,*SoundP;
    p=reader->MovieClipsPointer;
    while (p->NextP) {
        p=p->NextP;
    }
    SoundP=(struct  StartSound *)malloc(sizeof(struct  StartSound));
    bzero((void *)SoundP,sizeof(struct  StartSound));
    if (reader->read(reader, &(SoundP->SoundId), 2) !=2 )
        return gg_error();
    SoundP->SoundId=SWAP16(SoundP->SoundId);
    g_GetSoundInfo(reader,len,&(SoundP->SoundInfoP));
    Sp=p->MovieClipP->Frames[p->frame].EventSoundP;
    if (Sp) {
        while (Sp->NextP)
            Sp=Sp->NextP;
        Sp->NextP=SoundP;
    } else {
        p->MovieClipP->Frames[p->frame].EventSoundP = SoundP;
    }
    return 0;
}

int g_GetSoundStreamHead(struct reader_t *reader,int len) {
    struct MovieClips *MCp;
    struct SoundStreamHead *p;
    UI16 Num;
    SI16 SNum;
    MCp=reader->MovieClipsPointer;
    while (MCp->NextP) {
        MCp=MCp->NextP;
    }
    p=(struct SoundStreamHead*)malloc(sizeof(struct SoundStreamHead));
    bzero((void*)p,sizeof(struct SoundStreamHead));
    if (reader->read(reader, &(p->PlaybackFlags), 1) !=1 )
        return gg_error();
    if (reader->read(reader, &(p->StreamSoundFlags), 1) !=1 )
        return gg_error();
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->StreamSoundSampleCount=SWAP16(Num);
    if((p->StreamSoundFlags&MP3)==MP3) {
        if (reader->read(reader, &SNum, 2) !=2 )
            return gg_error();
        p->LatencySeek=SWAP16(SNum);
    }
    MCp->MovieClipP->Frames[MCp->frame].SoundStreamHeadP=p;
    return 0;
}

int g_GetRemoveObject(struct reader_t *reader,int len,int type) {
    struct RemoveObject *p;
    struct PAndRObjects *dq;
    UI16 Num;
    struct MovieClips *MCp;
    MCp=reader->MovieClipsPointer;
    while (MCp->NextP) {
        MCp=MCp->NextP;
    }
    p=(struct RemoveObject *)malloc(sizeof(struct RemoveObject));
    bzero((void *)p,sizeof(struct RemoveObject));
    if (type==1) {
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->CharacterID=SWAP16(Num);
        p->PlaceAndRemoveType=TRemoveObjectOne;
    } else {
        p->PlaceAndRemoveType=TRemoveObjectTwo;
    }
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->Depth=SWAP16(Num);
    dq=MCp->MovieClipP->Frames[MCp->frame].TagsListP;
    if (dq) {
        while (dq->NextP)
            dq=dq->NextP;
        dq->NextP=(struct PAndRObjects *)p;
    } else {
        MCp->MovieClipP->Frames[MCp->frame].TagsListP=(struct PAndRObjects *)p;
    }
    return 0;
}
void reader_AddActionRecord(struct ActionRecord** add_to_p,struct ActionRecord* ar_p) {
    struct ActionRecord* p = NULL;

    p = *add_to_p;
    if (p) {
        while (p->NextP) {
            p = p->NextP;
        };
        p->NextP=ar_p;
    } else {
        *add_to_p = ar_p;
    };
};

int reader_GetActionRecord(struct reader_t *reader,int len,struct ActionRecord **ActionRecordP) {
    struct ActionRecord *p,*dq;//,*decomp;
    //UI8 count=0;
    //UI16 ecount=0;
    UI32 length;
    //long i;
    length=len;
    p=(struct ActionRecord *)malloc(sizeof(struct ActionRecord)+length-1);
    bzero((void *)p,sizeof(struct ActionRecord)+length-1);
    p->Length=length;
    if (reader->read(reader, (p->Code), length) != length)
        return gg_error();
    //decomp_action(p,&decomp);
    //free(p);
    if(*ActionRecordP) {
        dq=*ActionRecordP;
        while(dq->NextP)
            dq=dq->NextP;
        dq->NextP=p;
    } else {
        *ActionRecordP=p;
    }
    return 0;
    /*    if(len==0){
    while(1){
    if (reader->read(reader, &count, 1) != 1) return 0;
    length=count;
    if (!count) return 0;
    if (count>=0x80){
    if (reader->read(reader, &ecount, 2) != 2) return 0;
    length=ecount;
    }
    p=(struct ActionRecord *)malloc(sizeof(struct ActionRecord)+length-1);
    bzero((void *)p,sizeof(struct ActionRecord)+length-1);
    p->Length=length;
    if (reader->read(reader, (p->Code), length) != length) return 0;
    if(*ActionRecordP){
    dq=*ActionRecordP;
    while(dq->NextP) dq=dq->NextP;
    dq->NextP=p;
    }else{
     *ActionRecordP=p;
     }
     }
     }else{
     i=len;
     while(i){
     if (reader->read(reader, &count, 1) != 1) return 0;
     i-=1;
     length=count;
     if (!count) break;
     if (count>=0x80){
     if (reader->read(reader, &ecount, 2) != 2) return 0;
     i-=2;
     length=ecount;
     }
     p=(struct ActionRecord *)malloc(sizeof(struct ActionRecord)+length-1);
     bzero((void *)p,sizeof(struct ActionRecord)+length-1);
     p->Length=length;
     if (reader->read(reader, (p->Code), length) != length) return 0;
     i-=length;
     if(*ActionRecordP){
     dq=*ActionRecordP;
     while(dq->NextP) dq=dq->NextP;
     dq->NextP=p;
     }else{
     *ActionRecordP=p;
     }
     if (i<0){
     printf("read Action Record error!\n");
     exit(0);
     }
     }
     }
     */
}

int g_GetPlaceObject2(struct reader_t *reader,swf_file_t *swf_p,int len) {
    struct PlaceObject *p;
    struct PAndRObjects *dq;
    char ptr[4];
    char *strp;
    UI16 Num;
    UI16 Flags16;
    //UI8 KeyCode;
    struct ActionRecord *ActionRecordP;
    UI32 ActionRecordSize;
    struct MovieClips *MCp;
    MCp=reader->MovieClipsPointer;
    while (MCp->NextP) {
        MCp=MCp->NextP;
    }
    p=(struct PlaceObject *)malloc(sizeof(struct PlaceObject));
    bzero((void *)p,sizeof(struct PlaceObject));
    p->PlaceAndRemoveType=TPlaceObjectTwo;
    if (reader->read(reader, &(p->Flags), 1) != 1)
        return gg_error();
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->Depth=SWAP16(Num);
    if ((p->Flags&PlaceFlagHasCharacter)==PlaceFlagHasCharacter) {
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->CharacterID=SWAP16(Num);
    }
    if ((p->Flags&PlaceFlagHasMatrix)==PlaceFlagHasMatrix) {
        reader_GetMatrix(reader,&(p->Matrix));
    }
    if ((p->Flags&PlaceFlagHasColorTransform)==PlaceFlagHasColorTransform) {
        reader_GetCxFromWithAlpha(reader,&(p->ColorTransform));
    }
    if ((p->Flags&PlaceFlagHasRatio)==PlaceFlagHasRatio) {
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->Ratio=SWAP16(Num);
    }
    if ((p->Flags&PlaceFlagHasName)==PlaceFlagHasName) {
        reader_GetString(reader,&strp);
        p->NameP=strp;
#if DEBUG

        printf("namep:%s\n",tmpmcp->NameP);
#endif

    }
    if ((p->Flags&PlaceFlagHasClipDepth)==PlaceFlagHasClipDepth) {
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->ClipDepth=SWAP16(Num);
    }
    if ((p->Flags&PlaceFlagHasClipActions)==PlaceFlagHasClipActions) {
        struct ActionBlock	*actions_p = g_new0(struct ActionBlock,1);
        p->ActionBlockP = actions_p;

        if (swf_p->version<=5) {
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            while(1) {
                if (reader->read(reader, &Flags16, 2) !=2 )
                    return gg_error();
              //  Flags16=SWAP16(Flags16);
		printf("v5 flags %d\n",Flags16);
                if (!Flags16)
                    break;
                if (reader->read(reader, &ptr, 4) !=4 )
                    return gg_error();
                ActionRecordSize=GET32(ptr);
                if(ActionRecordSize) {
                    ActionRecordP=NULL;
                    reader_GetActionRecord(reader,ActionRecordSize,&ActionRecordP);
                    if ((Flags16&(TClipEventKeyupP))==(TClipEventKeyupP)) {
                        reader_AddActionRecord(&actions_p->ClipEventKeyupP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventKeyDownP))==(TClipEventKeyDownP)) {
                        reader_AddActionRecord(&actions_p->ClipEventKeyDownP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventMouseUpP))==(TClipEventMouseUpP)) {
                        reader_AddActionRecord(&actions_p->ClipEventMouseUpP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventMouseDownP))==(TClipEventMouseDownP)) {
                        reader_AddActionRecord(&actions_p->ClipEventMouseDownP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventMouseMoveP))==(TClipEventMouseMoveP)) {
                        reader_AddActionRecord(&actions_p->ClipEventMouseMoveP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventUnloadP))==(TClipEventUnloadP)) {
                        reader_AddActionRecord(&actions_p->ClipEventUnloadP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventEnterFrameP))==(TClipEventEnterFrameP)) {
                        reader_AddActionRecord(&actions_p->ClipEventEnterFrameP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventLoadP))==(TClipEventLoadP)) {
                        reader_AddActionRecord(&actions_p->ClipEventLoadP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventDragOverP))==(TClipEventDragOverP)) {
                        reader_AddActionRecord(&actions_p->ClipEventDragOverP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventRollOutP))==(TClipEventRollOutP)) {
                        reader_AddActionRecord(&actions_p->ClipEventRollOutP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventRollOverP))==(TClipEventRollOverP)) {
                        reader_AddActionRecord(&actions_p->ClipEventRollOverP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventReleaseOutsideP))==(TClipEventReleaseOutsideP)) {
                        reader_AddActionRecord(&actions_p->ClipEventReleaseOutsideP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventReleaseP))==(TClipEventReleaseP)) {
                        reader_AddActionRecord(&actions_p->ClipEventReleaseP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventPressP))==(TClipEventPressP)) {
                        reader_AddActionRecord(&actions_p->ClipEventPressP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventInitializeP))==(TClipEventInitializeP)) {
                        reader_AddActionRecord(&actions_p->ClipEventInitializeP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventDataP))==(TClipEventDataP)) {
                        reader_AddActionRecord(&actions_p->ClipEventDataP,ActionRecordP);
                    }
                }
            }
        } else {
            //for v6 and newer
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            while(1) {
                UI16 Flagsv6;
                UI8 keycode;
                if (reader->read(reader, &Flags16, 2) !=2 )
                    return gg_error();
                Flags16=SWAP16(Flags16);
                if (reader->read(reader, &Flagsv6, 2) !=2 )
                    return gg_error();
                Flagsv6=SWAP16(Flagsv6);
                if (!(Flags16|Flagsv6))
                    break;
		printf("v6 flags %d\n",Flags16);
                if ((Flagsv6&(TClipEventKeyPressP))==(TClipEventKeyPressP)) {
                    if (reader->read(reader, &keycode, 1) !=1 )
                        return gg_error();
                }
                if (reader->read(reader, &ptr, 4) !=4 )
                    return gg_error();
                ActionRecordSize=GET32(ptr);
                if(ActionRecordSize) {
                    ActionRecordP=NULL;
                    reader_GetActionRecord(reader,ActionRecordSize,&ActionRecordP);
                    if ((Flags16&(TClipEventKeyupP))==(TClipEventKeyupP)) {
                        reader_AddActionRecord(&actions_p->ClipEventKeyupP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventKeyDownP))==(TClipEventKeyDownP)) {
                        reader_AddActionRecord(&actions_p->ClipEventKeyDownP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventMouseUpP))==(TClipEventMouseUpP)) {
                        reader_AddActionRecord(&actions_p->ClipEventMouseUpP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventMouseDownP))==(TClipEventMouseDownP)) {
                        reader_AddActionRecord(&actions_p->ClipEventMouseDownP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventMouseMoveP))==(TClipEventMouseMoveP)) {
                        reader_AddActionRecord(&actions_p->ClipEventMouseMoveP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventUnloadP))==(TClipEventUnloadP)) {
                        reader_AddActionRecord(&actions_p->ClipEventUnloadP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventEnterFrameP))==(TClipEventEnterFrameP)) {
                        printf("ClipEventEnterFrameP in split\n");
                        reader_AddActionRecord(&actions_p->ClipEventEnterFrameP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventLoadP))==(TClipEventLoadP)) {
                        reader_AddActionRecord(&actions_p->ClipEventLoadP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventDragOverP))==(TClipEventDragOverP)) {
                        reader_AddActionRecord(&actions_p->ClipEventDragOverP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventRollOutP))==(TClipEventRollOutP)) {
                        reader_AddActionRecord(&actions_p->ClipEventRollOutP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventRollOverP))==(TClipEventRollOverP)) {
                        reader_AddActionRecord(&actions_p->ClipEventRollOverP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventReleaseOutsideP))==(TClipEventReleaseOutsideP)) {
                        reader_AddActionRecord(&actions_p->ClipEventReleaseOutsideP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventReleaseP))==(TClipEventReleaseP)) {
                        reader_AddActionRecord(&actions_p->ClipEventReleaseP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventPressP))==(TClipEventPressP)) {
                        reader_AddActionRecord(&actions_p->ClipEventPressP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventInitializeP))==(TClipEventInitializeP)) {
                        reader_AddActionRecord(&actions_p->ClipEventInitializeP,ActionRecordP);
                    }
                    if ((Flags16&(TClipEventDataP))==(TClipEventDataP)) {
                        reader_AddActionRecord(&actions_p->ClipEventDataP,ActionRecordP);
                    }
                    if ((Flagsv6&(TClipEventConstructP))==(TClipEventConstructP)) {
                        reader_AddActionRecord(&actions_p->ClipEventConstructP,ActionRecordP);
                    }
                    if ((Flagsv6&(TClipEventDragOutP))==(TClipEventDragOutP)) {
                        reader_AddActionRecord(&actions_p->ClipEventDragOutP,ActionRecordP);
                    }
                    if ((Flagsv6&(TClipEventKeyPressP))==(TClipEventKeyPressP)) {
                        reader_AddActionRecord(&actions_p->ClipEventKeyPressP[keycode],ActionRecordP);
                    }
                }
            }
        }
    }
    dq=MCp->MovieClipP->Frames[MCp->frame].TagsListP;
    if (dq) {
        while (dq->NextP)
            dq=dq->NextP;
        dq->NextP=(struct PAndRObjects *)p;
    } else {
        MCp->MovieClipP->Frames[MCp->frame].TagsListP=(struct PAndRObjects *)p;
    }
    return 0;
}


int g_GetPlaceObject(struct reader_t *reader,int len) {
    struct PlaceObject *p;
    struct PAndRObjects *dq;
    UI16 Num;
    int pos;
    int i;
    struct MovieClips *MCp;
    pos=len;
    MCp=reader->MovieClipsPointer;
    while (MCp->NextP) {
        MCp=MCp->NextP;
    }
    p=(struct PlaceObject *)malloc(sizeof(struct PlaceObject));
    bzero((void *)p,sizeof(struct PlaceObject));
    p->PlaceAndRemoveType=TPlaceObjectOne;
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->CharacterID=SWAP16(Num);
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->Depth=SWAP16(Num);
    p->Flags|=PlaceFlagHasMatrix;
    p->Flags|=PlaceFlagHasCharacter;
    pos-=4;
    i=reader->pos;
    reader_GetMatrix(reader,&(p->Matrix));
    i-=reader->pos;
    pos+=i;
    if (pos>0) {
        p->Flags|=PlaceFlagHasColorTransform;
        reader_GetCxFrom(reader,&(p->ColorTransform));
    }
    dq=MCp->MovieClipP->Frames[MCp->frame].TagsListP;
    if (dq) {
        while (dq->NextP)
            dq=dq->NextP;
        dq->NextP=(struct PAndRObjects *)p;
    } else {
        MCp->MovieClipP->Frames[MCp->frame].TagsListP=(struct PAndRObjects *)p;
    }
    return 0;
}

int g_GetFrameLabel(struct reader_t *reader,int len) {
    struct FrameLabel *p,*dq;
    struct MovieClips *MCp;
    int size,i;
    UI8 tmp;
    size=len;
    p=(struct FrameLabel *)malloc(sizeof(struct FrameLabel));
    bzero((void *)p,sizeof(struct FrameLabel));
    MCp=reader->MovieClipsPointer;
    while (MCp->NextP) {
        MCp=MCp->NextP;
    }
    p->Frame=MCp->frame;
    size-=reader_GetString(reader,&(p->NameP));
    for(i=0;i<size;i++) {
        if (reader->read(reader, &tmp, 1) !=1 )
            return gg_error();
    }
    dq=MCp->MovieClipP->FrameLabelListP;
    if (dq) {
        while (dq->NextP)
            dq=dq->NextP;
        dq->NextP=p;
    } else {
        MCp->MovieClipP->FrameLabelListP=p;
    }
    return 0;
}

int g_GetEnd(struct reader_t *reader) {
    struct MovieClips *MCp,*pr=NULL;
    MCp=reader->MovieClipsPointer;
    while (MCp->NextP) {
        pr=MCp;
        MCp=MCp->NextP;
    }
    if(MCp==reader->MovieClipsPointer) {
        return -1;
    } else {
        free(MCp);
        pr->NextP=NULL;
        return 0;
    }
}

int g_GetDefineSprite(struct reader_t *reader,swf_file_t *swf_p,int len) {
    UI16 id,count;
    struct MovieClip *p;
    struct MovieClips *MCp;
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    id=SWAP16(id);
    if (reader->read(reader, &count, 2) !=2 )
        return gg_error();
    count=SWAP16(count);
    p=(struct MovieClip *)malloc(sizeof(struct MovieClip)+sizeof(struct Frame)*count);
    bzero((void *)p,sizeof(struct MovieClip)+sizeof(struct Frame)*count);
    g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
    p->CharacterType = CharMovieClip;
    p->TotalFrames=count;
    MCp=reader->MovieClipsPointer;
    if (MCp) {
        while(MCp->NextP)
            MCp=MCp->NextP;
        MCp->NextP=(struct MovieClips *)malloc(sizeof(struct MovieClips));
        bzero((void *)MCp->NextP,sizeof(struct MovieClips));
        MCp=MCp->NextP;
        MCp->frame=0;
        MCp->MovieClipP=p;
    }
    return 0;
}

int g_GetJPEGTables(struct reader_t *reader,swf_file_t *swf_p,int len) {
    struct JPEGTable *p;
    p=(struct JPEGTable *)malloc(sizeof(struct JPEGTable)+len-1);
    p->Count=len;
    if (reader->read(reader, (void *)(p->JPEGTables), len) !=len )
        return gg_error();
    swf_p->JPEGTableP=p;
    return 0;
}

int g_GetDefineButtonSound(struct reader_t *reader,swf_file_t *swf_p,int len) {

    struct  DefineButton *p;
    UI16 id;
    UI16 Num;
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    id=SWAP16(id);
    p=((struct  DefineButton *)swf_p->dict_p->dict[id]);
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->ButtonSoundChar0=SWAP16(Num);
    if(p->ButtonSoundChar0) {
        g_GetSoundInfo(reader,0,&(p->ButtonSoundInfo0));
    }
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->ButtonSoundChar1=SWAP16(Num);
    if(p->ButtonSoundChar1) {
        g_GetSoundInfo(reader,0,&(p->ButtonSoundInfo1));
    }
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->ButtonSoundChar2=SWAP16(Num);
    if(p->ButtonSoundChar2) {
        g_GetSoundInfo(reader,0,&(p->ButtonSoundInfo2));
    }
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->ButtonSoundChar3=SWAP16(Num);
    if(p->ButtonSoundChar3) {
        g_GetSoundInfo(reader,0,&(p->ButtonSoundInfo3));
    }
    return 0;
}

int g_GetSoundStreamBlock(struct reader_t *reader,int len) {
    struct SoundStreamBlock *p;
    struct MovieClips *MCp;
    MCp=reader->MovieClipsPointer;
    while(MCp->NextP)
        MCp=MCp->NextP;
    p=(struct SoundStreamBlock *)malloc(sizeof(struct SoundStreamBlock)+len-1);
    bzero((void *)p,sizeof(struct SoundStreamBlock)+len-1);
    p->Count=len;
    p->MovieClipP=MCp->MovieClipP;
    if (reader->read(reader, (void *)(p->StreamSoundData), len) !=len )
        return gg_error();
    MCp->MovieClipP->Frames[MCp->frame].SoundStreamBlockP=p;
    return 0;

}

int g_GetDefineBitsLossLess(struct reader_t *reader,swf_file_t *swf_p,int len,int type) {
    struct DefineBitsLossless2 *p;
    UI16 id;
    UI16 Num;
    UI8 format;
    UI8 size=0;
    int datacount=0;
    datacount=len-7;
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    id=SWAP16(id);
    if (reader->read(reader, &format, 1) !=1 )
        return gg_error();
    if (format==3) {
        datacount-=1;
    }
    p=(struct DefineBitsLossless2 *)malloc(sizeof(struct DefineBitsLossless2)+datacount);
    bzero((void *)p,sizeof(struct DefineBitsLossless2)+datacount);
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->BitmapWidth=SWAP16(Num);
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->BitmapHeight=SWAP16(Num);
    if (format==3) {
        if (reader->read(reader, &size, 1) !=1 )
            return gg_error();
    }
    if (type==1) {
        p->CharacterType=(CharBitsLosslessOne);
    } else {
        p->CharacterType=(CharBitsLosslessTwo);
    }
    p->BitmapColorTableSize=size;
    p->BitmapFormat=format;
    p->ZlibBitmapDataLength=datacount;
    if (reader->read(reader, (p->ZlibBitmapData), datacount) !=datacount )
        return gg_error();
    // bmp_makebmp(p);
    return g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
}

int g_GetButtonRecords(struct reader_t *reader,struct  DefineButton *p,int type) {
    struct PlaceObject *POp=NULL,*POp1=NULL;
    struct PAndRObjects* dq;
    UI8 Flags;
    UI16 Num;
    int size=reader->pos;
    while(1) {
        if (!POp) {
            POp=(struct PlaceObject *)malloc(sizeof(struct PlaceObject));
            bzero((void *)POp,sizeof(struct PlaceObject));
        }
        if (reader->read(reader, &Flags, 1) !=1 )
            return gg_error();
        if (!Flags) {
            size-=reader->pos;
            free(POp);
            return size;
        }
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        POp->CharacterID=SWAP16(Num);
        POp->PlaceAndRemoveType=TPlaceObjectTwo;
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        POp->Depth=SWAP16(Num);
        reader_GetMatrix(reader,&(POp->Matrix));
        if (type==2) {
            reader_GetCxFromWithAlpha(reader,&(POp->ColorTransform));
        }
        if ((Flags&(BtnStateUp))==(BtnStateUp)) {
            POp1=(struct PlaceObject *)malloc(sizeof(struct PlaceObject));
            bzero((void *)POp1,sizeof(struct PlaceObject));
            memcpy((void *)POp1,(void *)POp,sizeof(struct PlaceObject));
            dq=(struct PAndRObjects*)p->StateUpP;
            if(dq) {
                while(dq->NextP)
                    dq=dq->NextP;
                dq->NextP=(struct PAndRObjects*)POp1;
            } else {
                p->StateUpP=POp1;
            }
        }
        if ((Flags&(BtnStateOver))==(BtnStateOver)) {
            POp1=(struct PlaceObject *)malloc(sizeof(struct PlaceObject));
            bzero((void *)POp1,sizeof(struct PlaceObject));
            memcpy((void *)POp1,(void *)POp,sizeof(struct PlaceObject));
            dq=(struct PAndRObjects*)p->StateOverP;
            if(dq) {
                while(dq->NextP)
                    dq=dq->NextP;
                dq->NextP=(struct PAndRObjects*)POp1;
            } else {
                p->StateOverP=POp1;
            }
        }
        if ((Flags&(BtnStateDown))==(BtnStateDown)) {
            POp1=(struct PlaceObject *)malloc(sizeof(struct PlaceObject));
            bzero((void *)POp1,sizeof(struct PlaceObject));
            memcpy((void *)POp1,(void *)POp,sizeof(struct PlaceObject));
            dq=(struct PAndRObjects*)p->StateDownP;
            if(dq) {
                while(dq->NextP)
                    dq=dq->NextP;
                dq->NextP=(struct PAndRObjects*)POp1;
            } else {
                p->StateDownP=POp1;
            }
        }
        if ((Flags&(BtnStateHitTest))==(BtnStateHitTest)) {
            POp1=(struct PlaceObject *)malloc(sizeof(struct PlaceObject));
            bzero((void *)POp1,sizeof(struct PlaceObject));
            memcpy((void *)POp1,(void *)POp,sizeof(struct PlaceObject));
            dq=(struct PAndRObjects*)p->StateHitTestP;
            if(dq) {
                while(dq->NextP)
                    dq=dq->NextP;
                dq->NextP=(struct PAndRObjects*)POp1;
            } else {
                p->StateHitTestP=POp1;
            }
        }
    }
}

int g_GetDefineButton(struct reader_t *reader,swf_file_t *swf_p,int len) {
    struct  DefineButton *p;
    UI16 id;
    int actionsize;
    p=malloc(sizeof(struct  DefineButton));
    bzero((void *)p,sizeof(struct  DefineButton));
    p->CharacterType=CharButton;
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    id=SWAP16(id);
    p->ButtonID=id;
    actionsize=len-2;
    actionsize-=g_GetButtonRecords(reader,p,1);
    reader_GetActionRecord(reader,actionsize,&(p->CondOverDownToOverUpP));
    return g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
}

int g_GetDefineButton2(struct reader_t *reader,swf_file_t *swf_p,int len) {
    struct  DefineButton *p;
    struct ActionRecord *ARp;
    UI16 id;
    UI16 ActionOffset;
    UI16 CondActionSize;
    UI8 Flags1,Flags2,keycode;
    int tagsize;
    int endactionsize;
    tagsize=len;
    endactionsize=tagsize-3;
    p=malloc(sizeof(struct  DefineButton));
    bzero((void *)p,sizeof(struct  DefineButton));
    p->CharacterType=CharButton;
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    id=SWAP16(id);
    p->ButtonID=id;
    if (reader->read(reader, &(p->Flags), 1) !=1 )
        return gg_error();
    if (reader->read(reader, &ActionOffset, 2) !=2 )
        return gg_error();
    ActionOffset=SWAP16(ActionOffset);
    endactionsize-=ActionOffset;
    g_GetButtonRecords(reader,p,2);
    if (ActionOffset) {
        CondActionSize=ActionOffset;
        while(CondActionSize) {
            if (reader->read(reader, &CondActionSize, 2) !=2 )
                return gg_error();
            CondActionSize=SWAP16(CondActionSize);
            if (reader->read(reader, &Flags1, 1) !=1 )
                return gg_error();
            if (reader->read(reader, &Flags2, 1) !=1 )
                return gg_error();
            keycode=(Flags2>>1);
            ARp=NULL;
            if (CondActionSize) {
                endactionsize-=CondActionSize;
                reader_GetActionRecord(reader,(CondActionSize-4),&(ARp));
            } else {
                reader_GetActionRecord(reader,(endactionsize-4),&(ARp));
            }
            if((Flags1&(TCondIdleToOverDown))) {
                reader_AddActionRecord(&p->CondIdleToOverDownP,ARp);
            }
            if((Flags1&(TCondOutDownToIdle))) {
                reader_AddActionRecord(&p->CondOutDownToIdleP,ARp);
            }
            if((Flags1&(TCondOutDownToOverDown))) {
                reader_AddActionRecord(&p->CondOutDownToOverDownP,ARp);
            }
            if((Flags1&(TCondOverDownToOutDown))) {
                reader_AddActionRecord(&p->CondOverDownToOutDownP,ARp);
            }
            if((Flags1&(TCondOverDownToOverUp))) {
                reader_AddActionRecord(&p->CondOverDownToOverUpP,ARp);
            }
            if((Flags1&(TCondOverUpToOverDown))) {
                reader_AddActionRecord(&p->CondOverUpToOverDownP,ARp);
            }
            if((Flags1&(TCondOverUpToIdle))) {
                reader_AddActionRecord(&p->CondOverUpToIdleP,ARp);
            }
            if((Flags1&(TCondIdleToOverUp))) {
                reader_AddActionRecord(&p->CondIdleToOverUpP,ARp);
            }
            if(keycode) {
                reader_AddActionRecord(&p->CondKeyPressP[keycode],ARp);
            }
            if(Flags2&0x1) {
                reader_AddActionRecord(&p->CondOverDownToIdleP,ARp);
            }
        }

    }
    return g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
}

int g_GetDefineButtonCxform(struct reader_t *reader,swf_file_t *swf_p,int len) {
    struct CxFormWithAlpha* p;
    UI16 id;
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    id=SWAP16(id);
    p=&(((struct  DefineButton *)swf_p->dict_p->dict[id])->ColorTransform);
    reader_GetCxFrom(reader,p);
    return 0;
}

int g_GetDefineFont(struct reader_t *reader,swf_file_t *swf_p,int len) {
    struct  DefineFont *p;
    UI16 id;
    UI16 offset;
    int count=0,i,tmpint;
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    id=SWAP16(id);
    if (reader->read(reader, &offset, 2) !=2 )
        return gg_error();
    offset=SWAP16(offset);
    count=offset/2;
    p=(struct DefineFont *)malloc(sizeof(struct DefineFont)+(count-1)*sizeof(struct GlyphShape));
    bzero((void *)p,sizeof(struct DefineFont)+(count-1)*sizeof(struct GlyphShape));
    p->CharacterType=CharFontOne;
    p->NumGlyphs=count;
    for (i=1;i<count;i++) {
        if (reader->read(reader, &offset, 2) !=2 )
            return gg_error();
    }
    tmpint=sizeof(struct GlyphShape);
    for (i=0;i<count;i++) {
        g_GetShapeRecords(reader,&(p->GlyphShape[i].ShapeRecordP),0);
    }
    return g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
}

int g_GetDefineFont2(struct reader_t *reader,swf_file_t *swf_p,int len) {
    struct  DefineFont *p;
    struct GlyphShape *dq;
    struct KerningRecord *KRp;
    char ptr[4];
    UI8 FontNameLen;
    UI16 Num;
    UI8 Flags;
    UI8 LanguageCode;
    char *FontNameP;
    UI16 id;
    UI32 CodeTableOffset;
    UI16 count;
    int i;
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    id=SWAP16(id);
    if (reader->read(reader, &Flags, 1) !=1 )
        return gg_error();
    if (reader->read(reader, &LanguageCode, 1) !=1 )
        return gg_error();
    if (reader->read(reader, &FontNameLen, 1) !=1 )
        return gg_error();
    reader_GetStringWithLen(reader,&FontNameP,FontNameLen);
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    count=SWAP16(Num);
    p=(struct DefineFont *)malloc(sizeof(struct DefineFont)+(count-1)*sizeof(struct GlyphShape));
    bzero((void *)p,sizeof(struct DefineFont)+(count-1)*sizeof(struct GlyphShape));
    p->CharacterType=CharFontTwo;
    p->Flags=Flags;
    p->LanguageCode=LanguageCode;
    p->FontNameP=FontNameP;
    p->NumGlyphs=count;
    for (i=0;i<count;i++) {
        if(p->Flags&FontFlagsWideOffsets) {
            if (reader->read(reader, ptr, 4) !=4 )
                return gg_error();
        } else {
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
        }
    }
    if(p->Flags&FontFlagsWideOffsets) {
        if (reader->read(reader, ptr, 4) !=4 )
            return gg_error();
        CodeTableOffset=GET32(ptr);
    } else {
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        CodeTableOffset=SWAP16(Num);
    }
    dq=(struct GlyphShape *)p->GlyphShape;
    for (i=0;i<count;i++) {
        g_GetShapeRecords(reader,&(dq->ShapeRecordP),0);
        dq++;
    }
    dq=(struct GlyphShape *)p->GlyphShape;
    for (i=0;i<p->NumGlyphs;i++) {
        if(p->Flags&FontFlagsWideCodes) {
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            dq->Code=SWAP16(Num);
        } else {
            if (reader->read(reader, &(dq->Code), 1) !=1 )
                return gg_error();
        }
        dq++;
    }
    if(p->Flags&FontFlagsHasLayout) {
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->FontAscent=SWAP16(Num);
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->FontDescent=SWAP16(Num);
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->FontLeading=SWAP16(Num);
        dq=(struct GlyphShape *)p->GlyphShape;
        for (i=0;i<p->NumGlyphs;i++) {
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            dq->FontAdvance=SWAP16(Num);
            dq++;
        }
        dq=(struct GlyphShape *)p->GlyphShape;
        for (i=0;i<p->NumGlyphs;i++) {
            reader_GetRect(reader,&(dq->FontBounds));
            dq++;
        }
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->KerningCount=SWAP16(Num);
        if (p->KerningCount) {
            p->KerningTableP=(struct KerningRecord *)malloc(sizeof(struct KerningRecord));
            bzero((void *)p->KerningTableP,sizeof(struct KerningRecord));
            KRp=p->KerningTableP;
            if(p->Flags&FontFlagsWideCodes) {
                if (reader->read(reader, &Num, 2) !=2 )
                    return gg_error();
                KRp->FontKerningCode1=SWAP16(Num);
                if (reader->read(reader, &Num, 2) !=2 )
                    return gg_error();
                KRp->FontKerningCode2=SWAP16(Num);
            } else {
                if (reader->read(reader, &(KRp->FontKerningCode1), 1) !=1 )
                    return gg_error();
                if (reader->read(reader, &(KRp->FontKerningCode2), 1) !=1 )
                    return gg_error();
            }
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            KRp->FontKerningAdjustment=SWAP16(Num);
            for(i=1;i<p->KerningCount;i++) {
                KRp->NextP=(struct KerningRecord *)malloc(sizeof(struct KerningRecord));
                bzero((void *)KRp->NextP,sizeof(struct KerningRecord));
                KRp=KRp->NextP;
                if(p->Flags&FontFlagsWideCodes) {
                    if (reader->read(reader, &Num, 2) !=2 )
                        return gg_error();
                    KRp->FontKerningCode1=SWAP16(Num);
                    if (reader->read(reader, &Num, 2) !=2 )
                        return gg_error();
                    KRp->FontKerningCode2=SWAP16(Num);
                } else {
                    if (reader->read(reader, &(KRp->FontKerningCode1), 1) !=1 )
                        return gg_error();
                    if (reader->read(reader, &(KRp->FontKerningCode2), 1) !=1 )
                        return gg_error();
                }
                if (reader->read(reader, &Num, 2) !=2 )
                    return gg_error();
                KRp->FontKerningAdjustment=SWAP16(Num);
            }
        }
    }
    return g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
}

int g_GetDefineFontInfo(struct reader_t *reader,swf_file_t *swf_p,int len,int type) {
    struct  DefineFont *p;
    struct GlyphShape *dq;
    UI16 id;
    UI8 FontNameLen;
    UI16 Num;
    int i;
    if (reader->read(reader, &id, 2) !=2 )
        return gg_error();
    id=SWAP16(id);
    p=(struct DefineFont *)(swf_p->dict_p->dict[id]);
    if (reader->read(reader, &FontNameLen, 1) !=1 )
        return gg_error();
    reader_GetStringWithLen(reader,&(p->FontNameP),FontNameLen);
    if (reader->read(reader, &(p->Flags), 1) !=1 )
        return gg_error();
    if (type==2) {
        if (reader->read(reader, &(p->LanguageCode), 1) !=1 )
            return gg_error();
    }
    dq=(struct GlyphShape *)p->GlyphShape;
    for (i=0;i<p->NumGlyphs;i++) {
        if(p->Flags&FontInfoFlagsWideCodes) {
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            dq->Code=SWAP16(Num);
        } else {
            if (reader->read(reader, &(dq->Code), 1) !=1 )
                return gg_error();
        }
        dq++;
    }
    return 0;
}

int g_GetTextRecords(struct reader_t *reader,int type,struct TextRecord **TextRecordsP,int GlyphBits,int AdvanceBits) {
    struct TextRecord tmp_record;
    struct TextRecord *dq,*p,*headP;
    UI16 Num;
    UI16 FontID;
    UI16 TextHeight;
    int i;
    UI8 GlyphCount;
    headP=NULL;
    while(1) {
        bzero((void *)(&tmp_record),sizeof(struct TextRecord));
        if (reader->read(reader, &(tmp_record.Flags), 1) !=1 )
            return gg_error();
        if (!tmp_record.Flags)
            break;
        if (tmp_record.Flags&StyleFlagsHasFont) {
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            FontID=SWAP16(Num);
            tmp_record.FontID=FontID;
        } else {
            tmp_record.FontID=FontID;
        }
        if (tmp_record.Flags&StyleFlagsHasColor) {
            if (type==2) {
                reader_GetRgba(reader,&(tmp_record.TextColor));
            } else {
                reader_GetRgb(reader,&(tmp_record.TextColor));
            }
        }
        if (tmp_record.Flags&StyleFlagsHasXOffset) {
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            tmp_record.XOffset=SWAP16(Num);
        }
        if (tmp_record.Flags&StyleFlagsHasYOffset) {
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            tmp_record.YOffset=SWAP16(Num);
        }
        if (tmp_record.Flags&StyleFlagsHasFont) {
            if (reader->read(reader, &Num, 2) !=2 )
                return gg_error();
            TextHeight=SWAP16(Num);
            tmp_record.TextHeight=TextHeight;
        } else {
            tmp_record.TextHeight=TextHeight;
        }
        if (reader->read(reader, &(GlyphCount), 1) !=1 )
            return gg_error();
        tmp_record.GlyphCount=GlyphCount;
        p=(struct TextRecord *)malloc(sizeof(struct TextRecord)+(GlyphCount-1)*sizeof(struct  GlyphEntry));
        bzero((void *)p,sizeof(struct TextRecord)+(GlyphCount-1)*sizeof(struct  GlyphEntry));
        memcpy((void *)p,(void *)(&tmp_record),sizeof(struct TextRecord));
        for(i=0;i<GlyphCount;i++) {
            p->GlyphEntries[i].GlyphIndex=reader_GetBits(reader, GlyphBits);
            p->GlyphEntries[i].GlyphAdvance=reader_GetSBits(reader, AdvanceBits);
        }
        dq=headP;
        if (!headP) {
            headP=p;
        } else {
            while(dq->NextP)
                dq=dq->NextP;
            dq->NextP=p;
        }
    }
    *TextRecordsP=headP;
    return 0;
}

int g_GetDefineText(struct reader_t *reader,swf_file_t *swf_p,int len,int type) {
    struct DefineText *p;
    UI16 Num;
    UI16 id;
    int pos;
    p=(struct DefineText *)malloc(sizeof(struct DefineText));
    bzero((void *)p,sizeof(struct DefineText));
    pos=reader->pos;
    p->CharacterType=CharText;
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    id=SWAP16(Num);
    reader_GetRect(reader,&(p->TextBounds));
    reader_GetMatrix(reader,&(p->TextMatrix));
    if (reader->read(reader, &(p->GlyphBits), 1) !=1 )
        return gg_error();
    if (reader->read(reader, &(p->AdvanceBits), 1) !=1 )
        return gg_error();
    g_GetTextRecords(reader,type,&(p->TextRecordsP),p->GlyphBits,p->AdvanceBits);
    pos-=reader->pos;
    if (pos+len>0) {
        g_readdata(reader,pos+len);
        printf("read Define Text error,id:%d\n",id);
    }
    //    text_to_shape(p);
    return g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
}

int g_GetDefineEditText(struct reader_t *reader,swf_file_t *swf_p,int len) {
    struct DefineEditText *p;
    UI16 Num;
    UI16 id;
    p=(struct DefineEditText *)malloc(sizeof(struct DefineEditText));
    bzero((void *)p,sizeof(struct DefineEditText));
    p->CharacterType=CharEditText;
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    id=SWAP16(Num);
    reader_GetRect(reader,&(p->Bounds));
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->Flags=SWAP16(Num);
    if (p->Flags&EditTextHasFont) {
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->FontID=SWAP16(Num);
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->FontHeight=SWAP16(Num);
    }
    if (p->Flags&EditTextHasTextColor) {
        reader_GetRgba(reader,&(p->TextColor));
    }
    if (p->Flags&EditTextHasMaxLength) {
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->MaxLength=SWAP16(Num);
    }
    if (p->Flags&EditTextHasLayout) {
        if (reader->read(reader, &(p->Align), 1) !=1 )
            return gg_error();
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->LeftMargin=SWAP16(Num);
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->RightMargin=SWAP16(Num);
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->Indent=SWAP16(Num);
        if (reader->read(reader, &Num, 2) !=2 )
            return gg_error();
        p->Leading=SWAP16(Num);
    }
    reader_GetString(reader,&(p->VariableNameP));
    if (p->Flags&EditTextHasText) {
        reader_GetString(reader,&(p->InitialTextP));
    }
    return g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
}

int g_GetDoAction(struct reader_t *reader,int len) {
    struct ActionRecord *p;
    struct MovieClips *MCp;
    p=NULL;
    MCp=reader->MovieClipsPointer;
    while (MCp->NextP) {
        MCp=MCp->NextP;
    }
    reader_GetActionRecord(reader,len,&p);
    reader_AddActionRecord(&MCp->MovieClipP->Frames[MCp->frame].FrameEventP,p);
    return 0;

}

int g_GetGradientRecords(struct reader_t *reader,struct  MorphGradRecord **GradientP) {
    struct  MorphGradRecord *head=NULL;
    struct  MorphGradRecord *p=NULL;
    struct  MorphGradRecord *dq=NULL;
    UI8 count;
    int i;
    if (reader->read(reader, &count, 1) !=1 )
        return gg_error();
    for (i=0;i<count;i++) {
        p=(struct MorphGradRecord *)malloc(sizeof(struct MorphGradRecord));
        bzero((void *)p,sizeof(struct  MorphGradRecord));
        if (reader->read(reader, &(p->StartRatio), 1) !=1 )
            return gg_error();
        reader_GetRgba(reader,&(p->StartColor));
        if (reader->read(reader, &(p->EndRatio), 1) !=1 )
            return gg_error();
        reader_GetRgba(reader,&(p->EndColor));
        if(head) {
            dq=head;
            while(dq->NextP)
                dq=dq->NextP;
            dq->NextP=p;
        } else {
            head=p;
        }
    }
    *GradientP=head;
    return 0;
}

int g_GetMorphFillStyle(struct reader_t *reader,struct MorphFillStyle *MorphFillStyle_p) {
    struct MorphFillStyle *p;
    p=MorphFillStyle_p;
    if (reader->read(reader, &(p->FillStyleType), 1) !=1 )
        return gg_error();
    if (p->FillStyleType==0x00) {
        reader_GetRgba(reader,&(p->StartColor));
        reader_GetRgba(reader,&(p->EndColor));
    }
    if ((p->FillStyleType==0x10)||(p->FillStyleType==0x12)) {
        reader_GetMatrix(reader,&(p->StartGradientMatrix));
        reader_GetMatrix(reader,&(p->EndGradientMatrix));
        g_GetGradientRecords(reader,&(p->GradientP));
    }
    if ((p->FillStyleType==0x41)||(p->FillStyleType==0x42)||(p->FillStyleType==0x43)||(p->FillStyleType==0x40)) {
        if (reader->read(reader, &(p->BitmapId), 2) !=2 )
            return gg_error();
        p->BitmapId=SWAP16(p->BitmapId);
        reader_GetMatrix(reader,&(p->StartBitmapMatrix));
        reader_GetMatrix(reader,&(p->EndBitmapMatrix));
    }
    return 0;
}

int g_GetMorphFillStyleArray(struct reader_t *reader,struct MorphFillStyleArray **p) {
    UI8 count;
    UI16 e_count;
    int len;
    int i;
    if (reader->read(reader, &count, 1) !=1 )
        return gg_error();
    len=count;
    if (count==0xff) {
        if (reader->read(reader, &e_count, 2) !=2 )
            return gg_error();
        e_count=SWAP16(e_count);
        len=e_count;
    }
    *p=(struct MorphFillStyleArray *)malloc(sizeof(struct MorphFillStyleArray)+sizeof(struct MorphFillStyle)*(len-1));
    bzero((void *)*p,sizeof(struct MorphFillStyleArray)+sizeof(struct MorphFillStyle)*(len-1));
    (*p)->count=len;
    for(i=0;i<len;i++) {
        g_GetMorphFillStyle(reader,(struct MorphFillStyle *)&((*p)->MorphFillStyles[i]));
    }
    return 0;
}

int g_GetMorphLineStyle(struct reader_t *reader,struct MorphLineStyle *MorphLineStyle_p) {
    struct MorphLineStyle *p;
    UI16 Num;
    p=MorphLineStyle_p;
    bzero((void *)p,sizeof(struct MorphLineStyle));
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->StartWidth=SWAP16(Num);
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    p->EndWidth=SWAP16(Num);
    reader_GetRgba(reader,&(p->StartColor));
    reader_GetRgba(reader,&(p->EndColor));
    return 0;
}

int g_GetMorphLineStyleArray(struct reader_t *reader,struct MorphLineStyleArray **p) {
    UI8 count;
    UI16 e_count;
    int len;
    int i;
    if (reader->read(reader, &count, 1) !=1 )
        return gg_error();
    if (count==0xff) {
        if (reader->read(reader, &e_count, 2) !=2 )
            return gg_error();
        e_count=SWAP16(e_count);
        len=e_count;
    } else {
        len=count;
    }
    *p=(struct MorphLineStyleArray*)malloc(sizeof(struct MorphLineStyleArray)+sizeof(struct MorphLineStyle)*(len-1));
    bzero((void *)*p,sizeof(struct MorphLineStyleArray)+sizeof(struct MorphLineStyle)*(len-1));
    (*p)->count=len;
    for(i=0;i<len;i++) {
        g_GetMorphLineStyle(reader,&((*p)->MorphLineStyles[i]));
    }
    return 0;
}

int g_GetDefineMorphShape(struct reader_t *reader,swf_file_t *swf_p,int len) {
    struct  DefineMorphShape *p;
    UI16 Num;
    UI16 id;
    char strp[4];
    UI32 offset;
    int err;
    err=reader->pos;
    p=(struct DefineMorphShape *)malloc(sizeof(struct DefineMorphShape));
    bzero((void *)p,sizeof(struct DefineMorphShape));
    p->CharacterType=CharMorphShape;
    if (reader->read(reader, &Num, 2) !=2 )
        return gg_error();
    id=SWAP16(Num);
    reader_GetRect(reader,&(p->StartBounds));
    reader_GetRect(reader,&(p->EndBounds));
    if (reader->read(reader, strp, 4) !=4 )
        return gg_error();
    offset=GET32(strp);
    if (offset) {
        g_GetMorphFillStyleArray(reader,&(p->MorphFillStylesP));
        g_GetMorphLineStyleArray(reader,&(p->MorphLineStylesP));
        g_GetShapeRecords(reader,&(p->StartEdgesP),0);
    }
    g_GetShapeRecords(reader,&(p->EndEdgesP),0);
    err-=reader->pos;
    if (err+len>0) {
        printf("read Morph shape error!id:%d\n",id);
        g_readdata(reader,err+len);
        return 0;
    } else {
        return g_AddToDict(swf_p,id,(struct DefineCharacter*)p);
    }
}

int spliter_read_tags(struct spliter* spliter_p) {
    struct reader_t *r_p 	= spliter_p->current_reader;
    struct swf_file *s_p	= spliter_p->swf_p;
    UI16 			raw;
    UI32 			len;
    int 			id;
    int 			err=0;
    struct MovieClip 		*p;

    p=(struct MovieClip *)malloc(sizeof(struct MovieClip)+sizeof(struct Frame)*s_p->frame_count);
    bzero((void *)p,sizeof(struct MovieClip)+sizeof(struct Frame)*s_p->frame_count);
    p->CharacterType	= (CharMovieClip);
    p->TotalFrames	= s_p->frame_count;
    g_AddToDict(s_p,0,(struct DefineCharacter*)p);

    spliter_p->current_reader->MovieClipsPointer 		= g_new0(struct MovieClips,1);
    spliter_p->current_reader->MovieClipsPointer->frame		= 0;
    spliter_p->current_reader->MovieClipsPointer->MovieClipP	= p;

    while (r_p->pos < s_p->file_length) {
        if (r_p->read(r_p, &raw, 2) !=2 )
            return gg_error();

        raw = SWAP16(raw);
        len = raw&0x3f;
        id  = raw>>6;

        if (len==0x3f) {
            if (r_p->read(r_p, &len, 4) != 4)
                return gg_error();
            len = SWAP32(len);
        }

#if DEBUG
        printf("tag ID +%x- length: %ld\n",id,len);
        printf("r_p pos  %d\n",r_p->pos);
#endif

        if (id==ST_DEFINESPRITE)
            len = 2*sizeof(UI16);
        switch(id) {
        case ST_END: {
                err=g_GetEnd(r_p);
                if (err==0) {
                    break;
                } else {
    		    r_p->pos = s_p->file_length;
                    return -1;
                }
            }
        case ST_SHOWFRAME: {
                err=g_GetShowframe(r_p);
                break;
            }
        case ST_DEFINESHAPE: {
                err=g_GetDefineShape(r_p,s_p,len,1);
                break;
            }
        case ST_FREECHARACTER: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_PLACEOBJECT: {
                err=g_GetPlaceObject(r_p,len);
                break;
            }
        case ST_REMOVEOBJECT: {
                err=g_GetRemoveObject(r_p,len,1);
                break;
            }
        case ST_DEFINEBITS: {
                err=g_GetDefineBits(r_p,s_p,len);
                break;
            }
        case ST_DEFINEBUTTON: {
                err=g_GetDefineButton(r_p,s_p,len);
                break;
            }
        case ST_JPEGTABLES: {
                err=g_GetJPEGTables(r_p,s_p,len);
                break;
            }
        case ST_SETBACKGROUNDCOLOR: {
                err=g_GetColor(r_p,&s_p->BackgroundColor);
                break;
            }
        case ST_DEFINEFONT: {
                err=g_GetDefineFont(r_p,s_p,len);
                break;
            }
        case ST_DEFINETEXT: {
                err=g_GetDefineText(r_p,s_p,len,1);
                break;
            }
        case ST_DEFINEEDITTEXT: {
                err=g_GetDefineEditText(r_p,s_p,len);
                break;
            }
        case ST_DOACTION: {
                err=g_GetDoAction(r_p,len);
                break;
            }
        case ST_DEFINEFONTINFO: {
                err=g_GetDefineFontInfo(r_p,s_p,len,1);
                break;
            }
        case ST_DEFINESOUND: {
                err=g_GetDefSound(r_p,s_p,len);
                break;
            }
        case ST_STARTSOUND: {
                err=g_GetStartSound(r_p,len);
                break;
            }
        case ST_DEFINEBUTTONSOUND: {
                err=g_GetDefineButtonSound(r_p,s_p,len);
                break;
            }
        case ST_SOUNDSTREAMHEAD: {
                err=g_GetSoundStreamHead(r_p,len);
                break;
            }
        case ST_SOUNDSTREAMBLOCK: {
                err=g_GetSoundStreamBlock(r_p,len);
                break;
            }
        case ST_DEFINEBITSLOSSLESS: {
                err=g_GetDefineBitsLossLess(r_p,s_p,len,1);
                break;
            }
        case ST_DEFINEBITSJPEG2: {
                err=g_GetDefineBitsJPEG(r_p,s_p,len,2);
                break;
            }
        case ST_DEFINESHAPE2: {
                err=g_GetDefineShape(r_p,s_p,len,2);
                break;
            }
        case ST_DEFINEBUTTONCXFORM: {
                err=g_GetDefineButtonCxform(r_p,s_p,len);
                break;
            }
        case ST_PROTECT: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_PLACEOBJECT2: {
                err=g_GetPlaceObject2(r_p,s_p,len);
                break;
            }
        case ST_REMOVEOBJECT2: {
                err=g_GetRemoveObject(r_p,len,2);
                break;
            }
        case ST_DEFINESHAPE3: {
                err=g_GetDefineShape(r_p,s_p,len,3);
                break;
            }
        case ST_DEFINETEXT2: {
                err=g_GetDefineText(r_p,s_p,len,2);
                break;
            }
        case ST_DEFINEBUTTON2: {
                err=g_GetDefineButton2(r_p,s_p,len);
                break;
            }
        case ST_DEFINEBITSJPEG3: {
                err=g_GetDefineBitsJPEG(r_p,s_p,len,3);
                break;
            }
        case ST_DEFINEBITSLOSSLESS2: {
                err=g_GetDefineBitsLossLess(r_p,s_p,len,2);
                break;
            }
        case ST_DEFINESPRITE: {
                err=g_GetDefineSprite(r_p,s_p,len);
                break;
            }
        case ST_NAMECHARACTER: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_SERIALNUMBER: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_GENERATORTEXT: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_FRAMELABEL: {
                err=g_GetFrameLabel(r_p,len);
                break;
            }
        case ST_SOUNDSTREAMHEAD2: {
                err=g_GetSoundStreamHead(r_p,len);
                break;
            }
        case ST_DEFINEMORPHSHAPE: {
                err=g_GetDefineMorphShape(r_p,s_p,len);
                break;
            }
        case ST_DEFINEFONT2: {
                err=g_GetDefineFont2(r_p,s_p,len);
                break;
            }
        case ST_TEMPLATECOMMAND: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_GENERATOR3: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_EXTERNALFONT: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_EXPORTASSETS: {
                err=g_GetExportAssets(r_p,s_p,len);
                break;
            }
        case ST_IMPORTASSETS: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_ENABLEDEBUGGER: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_REFLEX: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_DOINITACTION: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_DEFINEMOVIE: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_DEFINEVIDEOSTREAM: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_VIDEOFRAME: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_DEFINEFONTINFO2: {
                err=g_GetDefineFontInfo(r_p,s_p,len,2);
                break;
            }
        case ST_FREEALL: {
                err=g_readdata(r_p,len);
                break;
            }
        case ST_MX4: {
                err=g_readdata(r_p,len);
                break;
            }
        default: {
                err=g_readdata(r_p,len);
                break;
            }
        }

    };
    r_p->pos = s_p->file_length;
    return err;
}

void spliter_read_head(struct spliter* spliter_p) {
    struct reader_t *r 		= spliter_p->current_reader;
    struct swf_file *s_p	= spliter_p->swf_p;
    char b[32];                         // read Header
    int len;

    assert((len = r->read(r,b,8))==8);

    assert(b[0]=='F' || b[0]=='C');
    assert(b[1]=='W');
    assert(b[2]=='S');

    s_p->version 	= b[3];
    s_p->compressed  	= (b[0]=='C')?1:0;
    s_p->file_length  	= GET32(&b[4]);

    if(s_p->compressed) {
        spliter_p->zlib_reader 		= g_new0(struct reader_t,1);
        reader_init_zlibinflate(spliter_p->zlib_reader, r);
        spliter_p->current_reader	= spliter_p->zlib_reader;
        r 				= spliter_p->zlib_reader;
    }

    reader_GetRect(r, &s_p->frame_size);
    r->read(r, &s_p->frame_rate, 2);
    s_p->frame_rate = SWAP16(s_p->frame_rate);
    s_p->frame_rate = s_p->frame_rate/256;
    r->read(r, &s_p->frame_count, 2);
    s_p->frame_count = SWAP16(s_p->frame_count);
}

spliter_t* spliter_new(swf_file_t *swf_p) {
    struct spliter *spliter_p 	= g_new0(struct spliter,1);

    spliter_p->file_reader 	= g_new0(struct reader_t,1);
    reader_init_filereader(spliter_p->file_reader, swf_p->handle);
    spliter_p->current_reader 	= spliter_p->file_reader;
    spliter_p->swf_p		= swf_p;

    return spliter_p;
};

void spliter_destory(struct spliter *spliter_p) {}
;
